# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import abc
import decimal
import os

from gabbi import fixture
import mock
from oslo_config import cfg
from oslo_config import fixture as conf_fixture
from oslo_db.sqlalchemy import utils
import oslo_messaging
from oslo_messaging import conffixture
from oslo_policy import opts as policy_opts
import six
from stevedore import driver
from stevedore import extension
import webob.dec
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from cloudkitty.api import app
from cloudkitty.api import middleware
from cloudkitty import db
from cloudkitty.db import api as ck_db_api
from cloudkitty import messaging
from cloudkitty import rating
from cloudkitty import storage
from cloudkitty.storage.sqlalchemy import models
from cloudkitty import tests
from cloudkitty import utils as ck_utils


INITIAL_TIMESTAMP = 1420070400


class UUIDFixture(fixture.GabbiFixture):
    def start_fixture(self):
        FAKE_UUID = '6c1b8a30-797f-4b7e-ad66-9879b79059fb'
        patcher = mock.patch(
            'oslo_utils.uuidutils.generate_uuid',
            return_value=FAKE_UUID)
        patcher.start()
        self.patcher = patcher

    def stop_fixture(self):
        self.patcher.stop()


@six.add_metaclass(abc.ABCMeta)
class BaseExtensionFixture(fixture.GabbiFixture):
    klass = None
    namespace = None
    stevedore_mgr = None
    assert_args = {}

    @abc.abstractmethod
    def setup_fake_modules(self):
        pass

    def start_fixture(self):
        fake_extensions = self.setup_fake_modules()
        self.mock = mock.patch(self.klass)
        fake_mgr = self.stevedore_mgr.make_test_instance(
            fake_extensions,
            self.namespace)
        self.patch = self.mock.start()
        self.patch.return_value = fake_mgr

    def stop_fixture(self):
        self.patch.assert_called_with(
            self.namespace,
            **self.assert_args)
        self.mock.stop()


class CollectorExtensionsFixture(BaseExtensionFixture):
    klass = 'stevedore.driver.DriverManager'
    namespace = 'cloudkitty.collector.backends'
    stevedore_mgr = driver.DriverManager
    assert_args = {
        'invoke_kwds': {'period': 3600},
        'invoke_on_load': True}

    def setup_fake_modules(self):
        def fake_metric(start,
                        end=None,
                        project_id=None,
                        q_filter=None):
            return None

        fake_module1 = tests.FakeCollectorModule()
        fake_module1.collector_name = 'fake1'
        fake_module1.get_compute = fake_metric
        fake_module2 = tests.FakeCollectorModule()
        fake_module2.collector_name = 'fake2'
        fake_module2.get_volume = fake_metric
        fake_module3 = tests.FakeCollectorModule()
        fake_module3.collector_name = 'fake3'
        fake_module3.get_compute = fake_metric
        fake_extensions = [
            extension.Extension(
                'fake1',
                'cloudkitty.tests.FakeCollectorModule1',
                None,
                fake_module1),
            extension.Extension(
                'fake2',
                'cloudkitty.tests.FakeCollectorModule2',
                None,
                fake_module2),
            extension.Extension(
                'fake3',
                'cloudkitty.tests.FakeCollectorModule3',
                None,
                fake_module3)]
        return fake_extensions[0]


class RatingModulesFixture(BaseExtensionFixture):
    klass = 'stevedore.extension.ExtensionManager'
    namespace = 'cloudkitty.rating.processors'
    stevedore_mgr = extension.ExtensionManager
    assert_args = {
        'invoke_on_load': True}

    def setup_fake_modules(self):
        class FakeConfigController(rating.RatingRestControllerBase):
            _custom_actions = {
                'test': ['GET']
            }

            @wsme_pecan.wsexpose(wtypes.text)
            def get_test(self):
                """Return the list of every mapping type available.

                """
                return 'OK'

        fake_module1 = tests.FakeRatingModule()
        fake_module1.module_name = 'fake1'
        fake_module1.set_priority(3)
        fake_module2 = tests.FakeRatingModule()
        fake_module2.module_name = 'fake2'
        fake_module2.config_controller = FakeConfigController
        fake_module2.set_priority(1)
        fake_module3 = tests.FakeRatingModule()
        fake_module3.module_name = 'fake3'
        fake_module3.set_priority(2)
        fake_extensions = [
            extension.Extension(
                'fake1',
                'cloudkitty.tests.FakeRatingModule1',
                None,
                fake_module1),
            extension.Extension(
                'fake2',
                'cloudkitty.tests.FakeRatingModule2',
                None,
                fake_module2),
            extension.Extension(
                'fake3',
                'cloudkitty.tests.FakeRatingModule3',
                None,
                fake_module3)]
        return fake_extensions


class ConfigFixture(fixture.GabbiFixture):
    auth_strategy = 'noauth'

    def start_fixture(self):
        self.conf = None
        conf = conf_fixture.Config().conf
        policy_opts.set_defaults(conf)
        msg_conf = conffixture.ConfFixture(conf)
        msg_conf.transport_driver = 'fake'
        conf.import_group('api', 'cloudkitty.api.app')
        conf.set_override('auth_strategy', self.auth_strategy)
        conf.set_override('connection', 'sqlite:///', 'database')
        conf.set_override('policy_file',
                          os.path.abspath('etc/cloudkitty/policy.json'),
                          group='oslo_policy')
        conf.set_override('api_paste_config',
                          os.path.abspath(
                              'cloudkitty/tests/gabbi/gabbi_paste.ini')
                          )
        conf.import_group('storage', 'cloudkitty.storage')
        conf.set_override('backend', 'sqlalchemy', 'storage')
        self.conf = conf
        self.conn = ck_db_api.get_instance()
        migration = self.conn.get_migration()
        migration.upgrade('head')

    def stop_fixture(self):
        if self.conf:
            self.conf.reset()
        db.get_engine().dispose()


class ConfigFixtureKeystoneAuth(ConfigFixture):
    auth_strategy = 'keystone'

    def start_fixture(self):
        # Mocking the middleware process_request which check for credentials
        # here, the only check done is that the hardcoded token is the one
        # send by the query. If not, 401, else 200.
        def _mock_proc_request(self, request):
            token = 'c93e3e31342e4e32ba201fd3d70878b5'
            http_code = 401
            if 'X-Auth-Token' in request.headers and \
               request.headers['X-Auth-Token'] == token:
                http_code = 200

            return webob.Response(
                status_code=http_code,
                content_type='application/json'
            )

        self._orig_func = middleware.auth_token.AuthProtocol.process_request
        middleware.auth_token.AuthProtocol.process_request = _mock_proc_request

        super(ConfigFixtureKeystoneAuth, self).start_fixture()

    def stop_fixture(self):
        super(ConfigFixtureKeystoneAuth, self).stop_fixture()
        middleware.auth_token.AuthProtocol.process_request = self._orig_func


class BaseFakeRPC(fixture.GabbiFixture):
    endpoint = None

    def start_fixture(self):
        messaging.setup()
        target = oslo_messaging.Target(topic='cloudkitty',
                                       server=cfg.CONF.host,
                                       version='1.0')
        endpoints = [
            self.endpoint()
        ]
        self.server = messaging.get_server(target, endpoints)
        self.server.start()

    def stop_fixture(self):
        self.server.stop()


class QuoteFakeRPC(BaseFakeRPC):
    class FakeRPCEndpoint(object):
        target = oslo_messaging.Target(namespace='rating',
                                       version='1.0')

        def quote(self, ctxt, res_data):
            return str(1.0)

    endpoint = FakeRPCEndpoint


class BaseStorageDataFixture(fixture.GabbiFixture):
    def create_fake_data(self, begin, end):
        data = [{
            "period": {
                "begin": begin,
                "end": end},
            "usage": {
                "compute": [
                    {
                        "desc": {
                            "dummy": True,
                            "fake_meta": 1.0},
                        "vol": {
                            "qty": 1,
                            "unit": "nothing"},
                        "rating": {
                            "price": decimal.Decimal('1.337')}}]}}, {
            "period": {
                "begin": begin,
                "end": end},
            "usage": {
                "image": [
                    {
                        "desc": {
                            "dummy": True,
                            "fake_meta": 1.0},
                        "vol": {
                            "qty": 1,
                            "unit": "nothing"},
                        "rating": {
                            "price": decimal.Decimal('0.121')}}]}}]
        return data

    def start_fixture(self):
        auth = mock.patch(
            'keystoneauth1.loading.load_auth_from_conf_options',
            return_value=dict())
        session = mock.patch(
            'keystoneauth1.loading.load_session_from_conf_options',
            return_value=dict())
        with auth:
            with session:
                self.storage = storage.get_storage()
        self.storage.init()
        self.initialize_data()

    def stop_fixture(self):
        model = models.RatedDataFrame
        session = db.get_session()
        q = utils.model_query(
            model,
            session)
        q.delete()


class StorageDataFixture(BaseStorageDataFixture):
    def initialize_data(self):
        nodata_duration = (24 * 3 + 12) * 3600
        tenant_list = ['8f82cc70-e50c-466e-8624-24bdea811375',
                       '7606a24a-b8ad-4ae0-be6c-3d7a41334a2e']
        for tenant in tenant_list:
            for i in range(INITIAL_TIMESTAMP,
                           INITIAL_TIMESTAMP + nodata_duration,
                           3600):
                self.storage.nodata(i, i + 3600, tenant)
        data_ts = INITIAL_TIMESTAMP + nodata_duration + 3600
        data_duration = (24 * 2 + 8) * 3600
        for i in range(data_ts,
                       data_ts + data_duration,
                       3600):
            data = self.create_fake_data(i, i + 3600)
            self.storage.append(data, tenant_list[0])
        half_duration = int(data_duration / 2)
        for i in range(data_ts,
                       data_ts + half_duration,
                       3600):
            data = self.create_fake_data(i, i + 3600)
            self.storage.append(data, tenant_list[1])
        for i in range(data_ts + half_duration + 3600,
                       data_ts + data_duration,
                       3600):
            self.storage.nodata(i, i + 3600, tenant_list[1])


class NowStorageDataFixture(BaseStorageDataFixture):
    def initialize_data(self):
        begin = ck_utils.get_month_start_timestamp()
        for i in range(begin,
                       begin + 3600 * 12,
                       3600):
            data = self.create_fake_data(i, i + 3600)
            self.storage.append(data,
                                '3d9a1b33-482f-42fd-aef9-b575a3da9369')


class CORSConfigFixture(fixture.GabbiFixture):
    """Inject mock configuration for the CORS middleware."""

    def start_fixture(self):
        # Here we monkeypatch GroupAttr.__getattr__, necessary because the
        # paste.ini method of initializing this middleware creates its own
        # ConfigOpts instance, bypassing the regular config fixture.

        def _mock_getattr(instance, key):
            if key != 'allowed_origin':
                return self._original_call_method(instance, key)
            return "http://valid.example.com"

        self._original_call_method = cfg.ConfigOpts.GroupAttr.__getattr__
        cfg.ConfigOpts.GroupAttr.__getattr__ = _mock_getattr

    def stop_fixture(self):
        """Remove the monkeypatch."""
        cfg.ConfigOpts.GroupAttr.__getattr__ = self._original_call_method


def setup_app():
    messaging.setup()
    # FIXME(sheeprine): Extension fixtures are interacting with transformers
    # loading, since collectors are not needed here we shunt them
    no_collector = mock.patch(
        'cloudkitty.collector.get_collector',
        return_value=None)
    with no_collector:
        return app.load_app()



# Copyright 2017 GohighSec.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from cloudkitty.common.policies import base

info_policies = [
    policy.DocumentedRuleDefault(
        name='info:list_services_info',
        check_str=base.UNPROTECTED,
        description='List available services information in Cloudkitty.',
        operations=[{'path': '/v1/info/services',
                     'method': 'LIST'}]),
    policy.DocumentedRuleDefault(
        name='info:get_service_info',
        check_str=base.UNPROTECTED,
        description='Get specified service information.',
        operations=[{'path': '/v1/info/services/{service_id}',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name='info:get_config',
        check_str=base.UNPROTECTED,
        description='Get current configuration in Cloudkitty.',
        operations=[{'path': '/v1/info/config',
                     'method': 'GET'}])
]


def list_rules():
    return info_policies



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import decimal

from cloudkitty import rating


class Noop(rating.RatingProcessorBase):

    module_name = "noop"
    description = 'Dummy test module.'

    @property
    def enabled(self):
        """Check if the module is enabled

        :returns: bool if module is enabled
        """
        return True

    @property
    def priority(self):
        return 1

    def reload_config(self):
        pass

    def process(self, data):
        for cur_data in data:
            cur_usage = cur_data['usage']
            for service in cur_usage:
                for entry in cur_usage[service]:
                    if 'rating' not in entry:
                        entry['rating'] = {'price': decimal.Decimal(0)}
        return data



# -*- coding: utf-8 -*-
# Copyright 2016 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import copy

from cloudkitty import tests
from cloudkitty.transformer import gnocchi

GNOCCHI_COMPUTE = {
    'id': '2f58a438-3169-11e6-b36c-bfe1fa3241fe',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'display_name': 'test',
    'flavor_id': '6aa7b1ce-317c-11e6-92d2-835668472674',
    'image_ref': 'http://fakeglance/c8ae2e38-316d-11e6-b19a-dbee663ddaee',
    'metrics': {}}

TRANS_COMPUTE = {
    'instance_id': '2f58a438-3169-11e6-b36c-bfe1fa3241fe',
    'resource_id': '2f58a438-3169-11e6-b36c-bfe1fa3241fe',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'name': 'test',
    'flavor_id': '6aa7b1ce-317c-11e6-92d2-835668472674',
    'image_id': 'c8ae2e38-316d-11e6-b19a-dbee663ddaee',
    'metrics': {}}

GNOCCHI_IMAGE = {
    'id': '2f58a438-3169-11e6-b36c-bfe1fa3241fe',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'container_format': 'bare',
    'disk_format': 'raw',
    'metrics': {}}

TRANS_IMAGE = {
    'resource_id': '2f58a438-3169-11e6-b36c-bfe1fa3241fe',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'container_format': 'bare',
    'disk_format': 'raw',
    'metrics': {}}

GNOCCHI_VOLUME = {
    'id': '17992d58-316f-11e6-9528-1379eed8ebe4',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'display_name': 'vol1',
    'volume_type': 'lvmdriver-1',
    'metrics': {}}

TRANS_VOLUME = {
    'resource_id': '17992d58-316f-11e6-9528-1379eed8ebe4',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'name': 'vol1',
    'volume_type': 'lvmdriver-1',
    'metrics': {}}

GNOCCHI_NETWORK = {
    'id': '02f8e84e-317d-11e6-ad23-af0423cd2a97',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'name': 'network1',
    'metrics': {}}

TRANS_NETWORK = {
    'resource_id': '02f8e84e-317d-11e6-ad23-af0423cd2a97',
    'project_id': '4480c638-3169-11e6-91de-a3bd3a7d3afb',
    'user_id': '576808d8-3169-11e6-992b-5f931fc671df',
    'name': 'network1',
    'metrics': {}}


class GnocchiTransformerTest(tests.TestCase):
    def test_strip_gnocchi_compute(self):
        resource = copy.deepcopy(GNOCCHI_COMPUTE)
        t_test = gnocchi.GnocchiTransformer()
        result = t_test.strip_resource_data('compute', resource)
        self.assertEqual(TRANS_COMPUTE, result)

    def test_strip_gnocchi_image(self):
        resource = copy.deepcopy(GNOCCHI_IMAGE)
        t_test = gnocchi.GnocchiTransformer()
        result = t_test.strip_resource_data('image', resource)
        self.assertEqual(TRANS_IMAGE, result)

    def test_strip_gnocchi_volume(self):
        resource = copy.deepcopy(GNOCCHI_VOLUME)
        t_test = gnocchi.GnocchiTransformer()
        result = t_test.strip_resource_data('volume', resource)
        self.assertEqual(TRANS_VOLUME, result)

    def test_strip_gnocchi_network(self):
        resource = copy.deepcopy(GNOCCHI_NETWORK)
        t_test = gnocchi.GnocchiTransformer()
        result = t_test.strip_resource_data('network', resource)
        self.assertEqual(TRANS_NETWORK, result)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_db.sqlalchemy import models
import sqlalchemy
from sqlalchemy.ext import declarative
from sqlalchemy import orm
from sqlalchemy import schema

from cloudkitty.common.db import models as ck_models

Base = ck_models.get_base()


class HashMapBase(models.ModelBase):
    __table_args__ = {'mysql_charset': "utf8",
                      'mysql_engine': "InnoDB"}
    fk_to_resolve = {}

    def save(self, session=None):
        from cloudkitty import db

        if session is None:
            session = db.get_session()

        super(HashMapBase, self).save(session=session)

    def as_dict(self):
        d = {}
        for c in self.__table__.columns:
            if c.name == 'id':
                continue
            d[c.name] = self[c.name]
        return d

    def _recursive_resolve(self, path):
        obj = self
        for attr in path.split('.'):
            if hasattr(obj, attr):
                obj = getattr(obj, attr)
            else:
                return None
        return obj

    def export_model(self):
        res = self.as_dict()
        for fk, mapping in self.fk_to_resolve.items():
            res[fk] = self._recursive_resolve(mapping)
        return res


class HashMapService(Base, HashMapBase):
    """A hashmap service.

    Used to describe a CloudKitty service such as compute or volume.
    """
    __tablename__ = 'hashmap_services'

    id = sqlalchemy.Column(
        sqlalchemy.Integer,
        primary_key=True)
    service_id = sqlalchemy.Column(
        sqlalchemy.String(36),
        nullable=False,
        unique=True)
    name = sqlalchemy.Column(
        sqlalchemy.String(255),
        nullable=False,
        unique=True)
    fields = orm.relationship(
        'HashMapField',
        backref=orm.backref(
            'service',
            lazy='immediate'))
    mappings = orm.relationship(
        'HashMapMapping',
        backref=orm.backref(
            'service',
            lazy='immediate'))
    thresholds = orm.relationship(
        'HashMapThreshold',
        backref=orm.backref(
            'service',
            lazy='immediate'))

    def __repr__(self):
        return ('<HashMapService[{uuid}]: '
                'service={service}>').format(
                    uuid=self.service_id,
                    service=self.name)


class HashMapField(Base, HashMapBase):
    """A hashmap field.

    Used to describe a service metadata such as flavor_id or image_id for
    compute.
    """
    __tablename__ = 'hashmap_fields'
    fk_to_resolve = {
        'service_id': 'service.service_id'}

    @declarative.declared_attr
    def __table_args__(cls):
        args = (
            schema.UniqueConstraint(
                'service_id',
                'name',
                name='uniq_field_per_service'),
            HashMapBase.__table_args__,)
        return args

    id = sqlalchemy.Column(
        sqlalchemy.Integer,
        primary_key=True)
    field_id = sqlalchemy.Column(
        sqlalchemy.String(36),
        nullable=False,
        unique=True)
    name = sqlalchemy.Column(
        sqlalchemy.String(255),
        nullable=False)
    service_id = sqlalchemy.Column(
        sqlalchemy.Integer,
        sqlalchemy.ForeignKey(
            'hashmap_services.id',
            ondelete='CASCADE'),
        nullable=False)
    mappings = orm.relationship(
        'HashMapMapping',
        backref=orm.backref(
            'field',
            lazy='immediate'))
    thresholds = orm.relationship(
        'HashMapThreshold',
        backref=orm.backref(
            'field',
            lazy='immediate'))

    def __repr__(self):
        return ('<HashMapField[{uuid}]: '
                'field={field}>').format(
                    uuid=self.field_id,
                    field=self.name)


class HashMapGroup(Base, HashMapBase):
    """A grouping of hashmap calculations.

    Used to group multiple mappings or thresholds into a single calculation.
    """
    __tablename__ = 'hashmap_groups'

    id = sqlalchemy.Column(
        sqlalchemy.Integer,
        primary_key=True)
    group_id = sqlalchemy.Column(
        sqlalchemy.String(36),
        nullable=False,
        unique=True)
    name = sqlalchemy.Column(
        sqlalchemy.String(255),
        nullable=False,
        unique=True)
    mappings = orm.relationship(
        'HashMapMapping',
        backref=orm.backref(
            'group',
            lazy='immediate'))
    thresholds = orm.relationship(
        'HashMapThreshold',
        backref=orm.backref(
            'group',
            lazy='immediate'))

    def __repr__(self):
        return ('<HashMapGroup[{uuid}]: '
                'name={name}>').format(
                    uuid=self.group_id,
                    name=self.name)


class HashMapMapping(Base, HashMapBase):
    """A mapping between a field or service, a value and a type.

    Used to model final equation.
    """
    __tablename__ = 'hashmap_mappings'
    fk_to_resolve = {
        'service_id': 'service.service_id',
        'field_id': 'field.field_id',
        'group_id': 'group.group_id'}

    @declarative.declared_attr
    def __table_args__(cls):
        args = (
            schema.UniqueConstraint(
                'value',
                'field_id',
                'tenant_id',
                name='uniq_field_mapping'),
            schema.UniqueConstraint(
                'value',
                'service_id',
                'tenant_id',
                name='uniq_service_mapping'),
            HashMapBase.__table_args__,)
        return args

    id = sqlalchemy.Column(
        sqlalchemy.Integer,
        primary_key=True)
    mapping_id = sqlalchemy.Column(
        sqlalchemy.String(36),
        nullable=False,
        unique=True)
    value = sqlalchemy.Column(
        sqlalchemy.String(255),
        nullable=True)
    cost = sqlalchemy.Column(
        sqlalchemy.Numeric(20, 8),
        nullable=False)
    map_type = sqlalchemy.Column(
        sqlalchemy.Enum(
            'flat',
            'rate',
            name='enum_map_type'),
        nullable=False)
    service_id = sqlalchemy.Column(
        sqlalchemy.Integer,
        sqlalchemy.ForeignKey(
            'hashmap_services.id',
            ondelete='CASCADE'),
        nullable=True)
    field_id = sqlalchemy.Column(
        sqlalchemy.Integer,
        sqlalchemy.ForeignKey(
            'hashmap_fields.id',
            ondelete='CASCADE'),
        nullable=True)
    group_id = sqlalchemy.Column(
        sqlalchemy.Integer,
        sqlalchemy.ForeignKey(
            'hashmap_groups.id',
            ondelete='SET NULL'),
        nullable=True)
    tenant_id = sqlalchemy.Column(
        sqlalchemy.String(36),
        nullable=True)

    def __repr__(self):
        return ('<HashMapMapping[{uuid}]: '
                'type={map_type} {value}={cost}, '
                'tenant={tenant}>').format(
                    uuid=self.mapping_id,
                    map_type=self.map_type,
                    value=self.value,
                    cost=self.cost,
                    tenant=self.tenant_id)


class HashMapThreshold(Base, HashMapBase):
    """A threshold matching a service or a field with a level and a type.

    Used to model final equation.
    """
    __tablename__ = 'hashmap_thresholds'
    fk_to_resolve = {
        'service_id': 'service.service_id',
        'field_id': 'field.field_id',
        'group_id': 'group.group_id'}

    @declarative.declared_attr
    def __table_args__(cls):
        args = (
            schema.UniqueConstraint(
                'level',
                'field_id',
                'tenant_id',
                name='uniq_field_threshold'),
            schema.UniqueConstraint(
                'level',
                'service_id',
                'tenant_id',
                name='uniq_service_threshold'),
            HashMapBase.__table_args__,)
        return args

    id = sqlalchemy.Column(
        sqlalchemy.Integer,
        primary_key=True)
    threshold_id = sqlalchemy.Column(
        sqlalchemy.String(36),
        nullable=False,
        unique=True)
    level = sqlalchemy.Column(
        sqlalchemy.Numeric(20, 8),
        nullable=True)
    cost = sqlalchemy.Column(
        sqlalchemy.Numeric(20, 8),
        nullable=False)
    map_type = sqlalchemy.Column(
        sqlalchemy.Enum(
            'flat',
            'rate',
            name='enum_hashmap_type'),
        nullable=False)
    service_id = sqlalchemy.Column(
        sqlalchemy.Integer,
        sqlalchemy.ForeignKey(
            'hashmap_services.id',
            ondelete='CASCADE'),
        nullable=True)
    field_id = sqlalchemy.Column(
        sqlalchemy.Integer,
        sqlalchemy.ForeignKey(
            'hashmap_fields.id',
            ondelete='CASCADE'),
        nullable=True)
    group_id = sqlalchemy.Column(
        sqlalchemy.Integer,
        sqlalchemy.ForeignKey(
            'hashmap_groups.id',
            ondelete='SET NULL'),
        nullable=True)
    tenant_id = sqlalchemy.Column(
        sqlalchemy.String(36),
        nullable=True)

    def __repr__(self):
        return ('<HashMapThreshold[{uuid}]: '
                'type={map_type} {level}={cost}, '
                'tenant={tenant}>').format(
                    uuid=self.threshold_id,
                    map_type=self.map_type,
                    level=self.level,
                    cost=self.cost,
                    tenant=self.tenant_id)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Added support for meta collector

Revision ID: 2ac2217dcbd9
Revises: 464e951dc3b8
Create Date: 2014-09-25 12:41:28.585333

"""

# revision identifiers, used by Alembic.
revision = '2ac2217dcbd9'
down_revision = '464e951dc3b8'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'service_to_collector_mappings',
        sa.Column('service', sa.String(length=255), nullable=False),
        sa.Column('collector', sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint('service'))



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import pecan
from pecan import rest
import six
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1.datamodels import collector as collector_models
from cloudkitty.common import policy
from cloudkitty.db import api as db_api


class MappingController(rest.RestController):
    """REST Controller managing service to collector mappings."""

    def __init__(self):
        self._db = db_api.get_instance().get_service_to_collector_mapping()

    @wsme_pecan.wsexpose(collector_models.ServiceToCollectorMapping,
                         wtypes.text)
    def get_one(self, service):
        """Return a service to collector mapping.

        :param service: Name of the service to filter on.
        """
        policy.authorize(pecan.request.context, 'collector:get_mapping', {})
        try:
            mapping = self._db.get_mapping(service)
            return collector_models.ServiceToCollectorMapping(
                **mapping.as_dict())
        except db_api.NoSuchMapping as e:
            pecan.abort(404, six.text_type(e))

    @wsme_pecan.wsexpose(collector_models.ServiceToCollectorMappingCollection,
                         wtypes.text)
    def get_all(self, collector=None):
        """Return the list of every services mapped to a collector.

        :param collector: Filter on the collector name.
        :return: Service to collector mappings collection.
        """
        policy.authorize(pecan.request.context, 'collector:list_mappings', {})
        mappings = [collector_models.ServiceToCollectorMapping(
            **mapping.as_dict())
            for mapping in self._db.list_mappings(collector)]
        return collector_models.ServiceToCollectorMappingCollection(
            mappings=mappings)

    @wsme_pecan.wsexpose(collector_models.ServiceToCollectorMapping,
                         wtypes.text,
                         wtypes.text)
    def post(self, collector, service):
        """Create a service to collector mapping.

        :param collector: Name of the collector to apply mapping on.
        :param service: Name of the service to apply mapping on.
        """
        policy.authorize(pecan.request.context, 'collector:manage_mapping', {})
        new_mapping = self._db.set_mapping(service, collector)
        return collector_models.ServiceToCollectorMapping(
            service=new_mapping.service,
            collector=new_mapping.collector)

    @wsme_pecan.wsexpose(None,
                         wtypes.text,
                         status_code=204)
    def delete(self, service):
        """Delete a service to collector mapping.

        :param service: Name of the service to filter on.
        """
        policy.authorize(pecan.request.context, 'collector:manage_mapping', {})
        try:
            self._db.delete_mapping(service)
        except db_api.NoSuchMapping as e:
            pecan.abort(404, six.text_type(e))


class CollectorStateController(rest.RestController):
    """REST Controller managing collector states."""

    def __init__(self):
        self._db = db_api.get_instance().get_module_info()

    @wsme_pecan.wsexpose(collector_models.CollectorInfos, wtypes.text)
    def get(self, name):
        """Query the enable state of a collector.

        :param name: Name of the collector.
        :return: State of the collector.
        """
        policy.authorize(pecan.request.context, 'collector:get_state', {})
        enabled = self._db.get_state('collector_{}'.format(name))
        collector = collector_models.CollectorInfos(name=name,
                                                    enabled=enabled)
        return collector

    @wsme_pecan.wsexpose(collector_models.CollectorInfos,
                         wtypes.text,
                         body=collector_models.CollectorInfos)
    def put(self, name, infos):
        """Set the enable state of a collector.

        :param name: Name of the collector.
        :param infos: New state informations of the collector.
        :return: State of the collector.
        """
        policy.authorize(pecan.request.context, 'collector:update_state', {})
        enabled = self._db.set_state('collector_{}'.format(name),
                                     infos.enabled)
        collector = collector_models.CollectorInfos(name=name,
                                                    enabled=enabled)
        return collector


class CollectorController(rest.RestController):
    """REST Controller managing collector modules."""

    mappings = MappingController()
    states = CollectorStateController()

    # FIXME(sheeprine): Stub function used to pass requests to subcontrollers
    @wsme_pecan.wsexpose(None)
    def get(self):
        "Unused function, hack to let pecan route requests to subcontrollers."
        return



# -*- coding: utf-8 -*-
# Copyright 2016 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from cloudkitty.tests import samples
from cloudkitty import transformer


class Transformer(transformer.BaseTransformer):
    compute_map = {
        'name': ['name', 'display_name'],
        'flavor': ['flavor', 'flavor.name', 'instance_type'],
        'vcpus': ['vcpus'],
        'memory': ['memory', 'memory_mb'],
        'image_id': ['image_id', 'image.id', 'image_meta.base_image_ref'],
        'availability_zone': [
            'availability_zone',
            'OS-EXT-AZ.availability_zone'],
    }
    volume_map = {
        'volume_id': ['volume_id'],
        'name': ['display_name'],
        'availability_zone': ['availability_zone'],
        'size': ['size'],
    }
    test_map = {'test': lambda x, y: 'ok'}

    def _strip_network(self, res_metadata):
        return {'test': 'ok'}


class TransformerMeta(Transformer):
    metadata_item = 'metadata'


class EmptyClass(object):
    pass


class ClassWithAttr(object):
    def __init__(self, items=samples.COMPUTE_METADATA):
        for key, val in items.items():
            setattr(self, key, val)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import json

from oslo_db.sqlalchemy import models
import sqlalchemy
from sqlalchemy.ext import declarative

from cloudkitty import utils as ck_utils

Base = declarative.declarative_base()


class RatedDataFrame(Base, models.ModelBase):
    """A rated data frame.

    """
    __table_args__ = {'mysql_charset': "utf8",
                      'mysql_engine': "InnoDB"}
    __tablename__ = 'rated_data_frames'

    id = sqlalchemy.Column(sqlalchemy.Integer,
                           primary_key=True)
    tenant_id = sqlalchemy.Column(sqlalchemy.String(32),
                                  nullable=True)
    begin = sqlalchemy.Column(sqlalchemy.DateTime,
                              nullable=False)
    end = sqlalchemy.Column(sqlalchemy.DateTime,
                            nullable=False)
    unit = sqlalchemy.Column(sqlalchemy.String(255),
                             nullable=False)
    qty = sqlalchemy.Column(sqlalchemy.Numeric(15, 5),
                            nullable=False)
    res_type = sqlalchemy.Column(sqlalchemy.String(255),
                                 nullable=False)
    rate = sqlalchemy.Column(sqlalchemy.Float(),
                             nullable=False)
    desc = sqlalchemy.Column(sqlalchemy.Text(),
                             nullable=False)

    def to_cloudkitty(self, collector=None):
        # Rating informations
        rating_dict = {}
        rating_dict['price'] = self.rate

        # Volume informations
        vol_dict = {}
        vol_dict['qty'] = self.qty.normalize()
        vol_dict['unit'] = self.unit
        res_dict = {}

        # Encapsulate informations in a resource dict
        res_dict['rating'] = rating_dict
        res_dict['desc'] = json.loads(self.desc)
        res_dict['vol'] = vol_dict
        res_dict['tenant_id'] = self.tenant_id

        # Add resource to the usage dict
        usage_dict = {}
        usage_dict[self.res_type] = [res_dict]

        # Time informations
        period_dict = {}
        period_dict['begin'] = ck_utils.dt2iso(self.begin)
        period_dict['end'] = ck_utils.dt2iso(self.end)

        # Add period to the resource informations
        ck_dict = {}
        ck_dict['period'] = period_dict
        ck_dict['usage'] = usage_dict
        return ck_dict



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Initial migration

Revision ID: 17fd1b237aa3
Revises: None
Create Date: 2014-10-10 11:28:08.645122

"""

# revision identifiers, used by Alembic.
revision = '17fd1b237aa3'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'rated_data_frames',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('begin', sa.DateTime(), nullable=False),
        sa.Column('end', sa.DateTime(), nullable=False),
        sa.Column('unit', sa.String(length=255), nullable=False),
        sa.Column('qty', sa.Numeric(), nullable=False),
        sa.Column('res_type', sa.String(length=255), nullable=False),
        sa.Column('rate', sa.Float(), nullable=False),
        sa.Column('desc', sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_charset='utf8',
        mysql_engine='InnoDB')



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Initial migration

Revision ID: 464e951dc3b8
Revises: None
Create Date: 2014-08-05 17:41:34.470183

"""

# revision identifiers, used by Alembic.
revision = '464e951dc3b8'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'states',
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('state', sa.BigInteger(), nullable=False),
        sa.Column('s_metadata', sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint('name'))
    op.create_table(
        'modules_state',
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('state', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('name'))



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import socket
import sys

from oslo_config import cfg
import oslo_i18n
from oslo_log import log

from cloudkitty.common import defaults
from cloudkitty import messaging
from cloudkitty import version


service_opts = [
    cfg.StrOpt('host',
               default=socket.getfqdn(),
               sample_default='<server-hostname.example.com>',
               help='Name of this node. This can be an opaque identifier. '
               'It is not necessarily a hostname, FQDN, or IP address. '
               'However, the node name must be valid within an AMQP key, '
               'and if using ZeroMQ, a valid hostname, FQDN, or IP address.')
]

cfg.CONF.register_opts(service_opts)


def prepare_service(argv=None, config_files=None):
    oslo_i18n.enable_lazy()
    log.register_options(cfg.CONF)
    log.set_defaults()
    defaults.set_cors_middleware_defaults()

    if argv is None:
        argv = sys.argv
    cfg.CONF(argv[1:], project='cloudkitty', validate_default_values=True,
             version=version.version_info.version_string(),
             default_config_files=config_files)

    log.setup(cfg.CONF, 'cloudkitty')
    messaging.setup()



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import abc

import six
from stevedore import extension

TRANSFORMERS_NAMESPACE = 'cloudkitty.transformers'


def get_transformers():
    transformers = {}
    transformer_exts = extension.ExtensionManager(
        TRANSFORMERS_NAMESPACE,
        invoke_on_load=True)
    for transformer in transformer_exts:
        t_name = transformer.name
        t_obj = transformer.obj
        transformers[t_name] = t_obj
    return transformers


@six.add_metaclass(abc.ABCMeta)
class BaseTransformer(object):
    metadata_item = ''

    def generic_strip(self, datatype, data):
        metadata = getattr(data, self.metadata_item, data)
        mappings = getattr(self, datatype + '_map', {})
        result = {}
        for key, transform in mappings.items():
            if isinstance(transform, list):
                for meta_key in transform:
                    if key not in result or result[key] is None:
                        try:
                            data = getattr(metadata, meta_key)
                        except AttributeError:
                            data = metadata.get(meta_key)
                        result[key] = data
            else:
                trans_data = transform(self, metadata)
                if trans_data:
                    result[key] = trans_data
        return result

    def strip_resource_data(self, res_type, res_data):
        res_type = res_type.replace('.', '_')
        strip_func = getattr(self, '_strip_' + res_type, None)
        if strip_func:
            return strip_func(res_data)
        return self.generic_strip(res_type, res_data) or res_data

    def get_metadata(self, res_type):
        """Return list of metadata available for given resource type."""

        return []



# -*- coding: utf-8 -*-
# Copyright 2017 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Luka Peschke
#
from oslo_config import cfg
from oslo_db.sqlalchemy import utils
from stevedore import driver

from cloudkitty import db
from cloudkitty import storage
from cloudkitty.storage.hybrid import migration
from cloudkitty.storage.hybrid import models
from cloudkitty import utils as ck_utils


storage_opts = [
    cfg.StrOpt(
        'backend',
        default='gnocchi',
        help='Name of the storage backend that should be used '
        'by the hybrid storage')
]

CONF = cfg.CONF
CONF.register_opts(storage_opts, group='hybrid_storage')

HYBRID_BACKENDS_NAMESPACE = 'cloudkitty.storage.hybrid.backends'


class HybridStorage(storage.BaseStorage):
    """Hybrid Storage Backend.

    Stores dataframes in one of the available backends and other informations
    in a classical SQL database.
    """

    state_model = models.TenantState

    def __init__(self, **kwargs):
        super(HybridStorage, self).__init__(**kwargs)
        self._hybrid_backend = driver.DriverManager(
            HYBRID_BACKENDS_NAMESPACE,
            cfg.CONF.hybrid_storage.backend,
            invoke_on_load=True).driver
        self._sql_session = {}

    def _check_session(self, tenant_id):
        session = self._sql_session.get(tenant_id, None)
        if not session:
            self._sql_session[tenant_id] = db.get_session()
            self._sql_session[tenant_id].begin()

    def init(self):
        migration.upgrade('head')
        self._hybrid_backend.init()

    def get_state(self, tenant_id=None):
        session = db.get_session()
        q = utils.model_query(self.state_model, session)
        if tenant_id:
            q = q.filter(self.state_model.tenant_id == tenant_id)
        q = q.order_by(self.state_model.state.desc())
        r = q.first()
        return ck_utils.dt2ts(r.state) if r else None

    def _set_state(self, tenant_id, state):
        self._check_session(tenant_id)
        session = self._sql_session[tenant_id]
        q = utils.model_query(self.state_model, session)
        if tenant_id:
            q = q.filter(self.state_model.tenant_id == tenant_id)
        r = q.first()
        do_commit = False
        if r:
            if state > r.state:
                q.update({'state': state})
                do_commit = True
        else:
            state = self.state_model(tenant_id=tenant_id, state=state)
            session.add(state)
            do_commit = True
        if do_commit:
            session.commit()

    def _commit(self, tenant_id):
        self._hybrid_backend.commit(tenant_id, self.get_state(tenant_id))

    def _pre_commit(self, tenant_id):
        super(HybridStorage, self)._pre_commit(tenant_id)

    def _post_commit(self, tenant_id):
        self._set_state(tenant_id, self.usage_start_dt.get(tenant_id))
        super(HybridStorage, self)._post_commit(tenant_id)
        del self._sql_session[tenant_id]

    def get_total(self, begin=None, end=None, tenant_id=None,
                  service=None, groupby=None):
        return self._hybrid_backend.get_total(
            begin=begin, end=end, tenant_id=tenant_id,
            service=service, groupby=groupby)

    def _dispatch(self, data, tenant_id):
        if not self.get_state(tenant_id):
            self._set_state(tenant_id, self.usage_start_dt.get(tenant_id))
        for service in data:
            for frame in data[service]:
                self._hybrid_backend.append_time_frame(
                    service, frame, tenant_id)
                self._has_data[tenant_id] = True

    def get_tenants(self, begin, end):
        return self._hybrid_backend.get_tenants(begin, end)

    def get_time_frame(self, begin, end, **filters):
        return self._hybrid_backend.get_time_frame(begin, end, **filters)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Rename mapping table to hashmap_mappings.

Revision ID: 10d2738b67df
Revises: 54cc17accf2c
Create Date: 2016-05-24 18:37:25.305430

"""

# revision identifiers, used by Alembic.
revision = '10d2738b67df'
down_revision = '54cc17accf2c'

from alembic import op


def upgrade():
    op.rename_table('hashmap_maps', 'hashmap_mappings')



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_context import context
from pecan import hooks

from cloudkitty.common import policy
from cloudkitty import messaging


class RPCHook(hooks.PecanHook):
    def __init__(self):
        self._rpc_client = messaging.get_client()

    def before(self, state):
        state.request.rpc_client = self._rpc_client


class StorageHook(hooks.PecanHook):
    def __init__(self, storage_backend):
        self._storage_backend = storage_backend

    def before(self, state):
        state.request.storage_backend = self._storage_backend


class ContextHook(hooks.PecanHook):
    def on_route(self, state):
        headers = state.request.headers

        roles = headers.get('X-Roles', '').split(',')
        is_admin = policy.check_is_admin(roles)

        creds = {
            'user': headers.get('X-User-Id', ''),
            'tenant': headers.get('X-Tenant-Id', ''),
            'auth_token': headers.get('X-Auth-Token', ''),
            'is_admin': is_admin,
            'roles': roles,
            "user_name": headers.get('X-User-Name', ''),
            "project_name": headers.get('X-Project-Name', ''),
            "domain": headers.get('X-User-Domain-Id', ''),
            "domain_name": headers.get('X-User-Domain-Name', ''),
        }

        state.request.context = context.RequestContext(**creds)



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import pecan
import six
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1 import types as ck_types
from cloudkitty import rating
from cloudkitty.rating.hash.datamodels import group as group_models
from cloudkitty.rating.hash.datamodels import mapping as mapping_models
from cloudkitty.rating.hash.db import api as db_api


class HashMapMappingsController(rating.RatingRestControllerBase):
    """Controller responsible of mappings management.

    """

    _custom_actions = {
        'group': ['GET']}

    @wsme_pecan.wsexpose(group_models.Group,
                         ck_types.UuidType())
    def group(self, mapping_id):
        """Get the group attached to the mapping.

        :param mapping_id: UUID of the mapping to filter on.
        """
        hashmap = db_api.get_instance()
        try:
            group_db = hashmap.get_group_from_mapping(
                uuid=mapping_id)
            return group_models.Group(**group_db.export_model())
        except db_api.MappingHasNoGroup as e:
            pecan.abort(404, six.text_type(e))

    @wsme_pecan.wsexpose(mapping_models.MappingCollection,
                         ck_types.UuidType(),
                         ck_types.UuidType(),
                         ck_types.UuidType(),
                         bool,
                         ck_types.UuidType(),
                         bool,
                         status_code=200)
    def get_all(self,
                service_id=None,
                field_id=None,
                group_id=None,
                no_group=False,
                tenant_id=None,
                filter_tenant=False):
        """Get the mapping list

        :param service_id: Service UUID to filter on.
        :param field_id: Field UUID to filter on.
        :param group_id: Group UUID to filter on.
        :param no_group: Filter on orphaned mappings.
        :param tenant_id: Tenant UUID to filter on.
        :param filter_tenant: Explicitly filter on tenant (default is to not
        filter on tenant). Useful if you want to filter on tenant being None.
        :return: List of every mappings.
        """
        hashmap = db_api.get_instance()
        mapping_list = []
        search_opts = dict()
        if filter_tenant:
            search_opts['tenant_uuid'] = tenant_id
        mappings_uuid_list = hashmap.list_mappings(
            service_uuid=service_id,
            field_uuid=field_id,
            group_uuid=group_id,
            **search_opts)
        for mapping_uuid in mappings_uuid_list:
            mapping_db = hashmap.get_mapping(uuid=mapping_uuid)
            mapping_list.append(mapping_models.Mapping(
                **mapping_db.export_model()))
        res = mapping_models.MappingCollection(mappings=mapping_list)
        return res

    @wsme_pecan.wsexpose(mapping_models.Mapping,
                         ck_types.UuidType())
    def get_one(self, mapping_id):
        """Return a mapping.

        :param mapping_id: UUID of the mapping to filter on.
        """
        hashmap = db_api.get_instance()
        try:
            mapping_db = hashmap.get_mapping(uuid=mapping_id)
            return mapping_models.Mapping(
                **mapping_db.export_model())
        except db_api.NoSuchMapping as e:
            pecan.abort(404, six.text_type(e))

    @wsme_pecan.wsexpose(mapping_models.Mapping,
                         body=mapping_models.Mapping,
                         status_code=201)
    def post(self, mapping_data):
        """Create a mapping.

        :param mapping_data: Informations about the mapping to create.
        """
        hashmap = db_api.get_instance()
        try:
            mapping_db = hashmap.create_mapping(
                value=mapping_data.value,
                map_type=mapping_data.map_type,
                cost=mapping_data.cost,
                field_id=mapping_data.field_id,
                group_id=mapping_data.group_id,
                service_id=mapping_data.service_id,
                tenant_id=mapping_data.tenant_id)
            pecan.response.location = pecan.request.path_url
            if pecan.response.location[-1] != '/':
                pecan.response.location += '/'
            pecan.response.location += mapping_db.mapping_id
            return mapping_models.Mapping(
                **mapping_db.export_model())
        except db_api.MappingAlreadyExists as e:
            pecan.abort(409, six.text_type(e))
        except db_api.ClientHashMapError as e:
            pecan.abort(400, six.text_type(e))

    @wsme_pecan.wsexpose(None,
                         ck_types.UuidType(),
                         body=mapping_models.Mapping,
                         status_code=302)
    def put(self, mapping_id, mapping):
        """Update a mapping.

        :param mapping_id: UUID of the mapping to update.
        :param mapping: Mapping data to insert.
        """
        hashmap = db_api.get_instance()
        try:
            hashmap.update_mapping(
                mapping_id,
                mapping_id=mapping.mapping_id,
                value=mapping.value,
                cost=mapping.cost,
                map_type=mapping.map_type,
                group_id=mapping.group_id,
                tenant_id=mapping.tenant_id)
            pecan.response.headers['Location'] = pecan.request.path
        except db_api.NoSuchMapping as e:
            pecan.abort(404, six.text_type(e))
        except db_api.ClientHashMapError as e:
            pecan.abort(400, six.text_type(e))

    @wsme_pecan.wsexpose(None,
                         ck_types.UuidType(),
                         status_code=204)
    def delete(self, mapping_id):
        """Delete a mapping.

        :param mapping_id: UUID of the mapping to delete.
        """
        hashmap = db_api.get_instance()
        try:
            hashmap.delete_mapping(uuid=mapping_id)
        except db_api.NoSuchMapping as e:
            pecan.abort(404, six.text_type(e))



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Fixed constraint name.

Revision ID: 54cc17accf2c
Revises: 4fa888fd7eda
Create Date: 2015-05-28 16:44:32.936076

"""

# revision identifiers, used by Alembic.
revision = '54cc17accf2c'
down_revision = '4fa888fd7eda'

from alembic import op
import sqlalchemy as sa


def create_table(is_old=False):
    if is_old:
        constraints = ['uniq_field_mapping', 'uniq_service_mapping']
    else:
        constraints = ['uniq_field_threshold', 'uniq_service_threshold']
    table = op.create_table(
        'tmig_hashmap_thresholds',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('threshold_id', sa.String(length=36), nullable=False),
        sa.Column('level', sa.Numeric(precision=20, scale=8), nullable=True),
        sa.Column('cost', sa.Numeric(precision=20, scale=8), nullable=False),
        sa.Column(
            'map_type',
            sa.Enum('flat', 'rate', name='enum_map_type'),
            nullable=False),
        sa.Column('service_id', sa.Integer(), nullable=True),
        sa.Column('field_id', sa.Integer(), nullable=True),
        sa.Column('group_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['field_id'],
            ['hashmap_fields.id'],
            ondelete='CASCADE'),
        sa.ForeignKeyConstraint(
            ['group_id'],
            ['hashmap_groups.id'],
            ondelete='SET NULL'),
        sa.ForeignKeyConstraint(
            ['service_id'],
            ['hashmap_services.id'],
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('threshold_id'),
        sa.UniqueConstraint('level', 'field_id', name=constraints[0]),
        sa.UniqueConstraint('level', 'service_id', name=constraints[1]))
    return table


def upgrade():
    dialect = op.get_context().dialect.name
    try:
        # Needs sqlalchemy 0.8
        if dialect != 'postgresql':
            with op.batch_alter_table('hashmap_thresholds') as batch_op:
                batch_op.drop_constraint(
                    u'uniq_field_mapping',
                    type_='unique')
                batch_op.drop_constraint(
                    u'uniq_service_mapping',
                    type_='unique')
                batch_op.create_unique_constraint(
                    'uniq_field_threshold',
                    ['level', 'field_id'])
                batch_op.create_unique_constraint(
                    'uniq_service_threshold',
                    ['level', 'service_id'])
    except AttributeError:
        # No support for batch operations
        if dialect == 'sqlite':
            new_table = create_table()
            sel = sa.sql.expression.select(new_table.columns.keys())
            op.execute(
                new_table.insert().from_select(
                    new_table.columns.keys(),
                    sel.select_from('hashmap_thresholds')))
            op.drop_table('hashmap_thresholds')
            op.rename_table('tmig_hashmap_thresholds', 'hashmap_thresholds')
        else:
            op.drop_constraint(
                u'uniq_field_mapping',
                'hashmap_thresholds',
                type_='unique')
            op.drop_constraint(
                u'uniq_service_mapping',
                'hashmap_thresholds',
                type_='unique')
            op.create_unique_constraint(
                'uniq_field_threshold',
                'hashmap_thresholds',
                ['level', 'field_id'])
            op.create_unique_constraint(
                'uniq_service_threshold',
                'hashmap_thresholds',
                ['level', 'service_id'])



# Copyright 2017 GohighSec.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from cloudkitty.common.policies import base

rating_policies = [
    policy.DocumentedRuleDefault(
        name='rating:list_modules',
        check_str=base.ROLE_ADMIN,
        description='Reture the list of loaded modules in Cloudkitty.',
        operations=[{'path': '/v1/rating/modules',
                     'method': 'LIST'}]),
    policy.DocumentedRuleDefault(
        name='rating:get_module',
        check_str=base.ROLE_ADMIN,
        description='Get specified module.',
        operations=[{'path': '/v1/rating/modules/{module_id}',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name='rating:update_module',
        check_str=base.ROLE_ADMIN,
        description='Change the state and priority of a module.',
        operations=[{'path': '/v1/rating/modules/{module_id}',
                     'method': 'PUT'}]),
    policy.DocumentedRuleDefault(
        name='rating:quote',
        check_str=base.UNPROTECTED,
        description='Get an instant quote based on multiple resource '
                    'descriptions.',
        operations=[{'path': '/v1/rating/quote',
                     'method': 'POST'}]),
    policy.DocumentedRuleDefault(
        name='rating:module_config',
        check_str=base.ROLE_ADMIN,
        description='Trigger a rating module list reload.',
        operations=[{'path': '/v1/rating/reload_modules',
                     'method': 'GET'}])
]


def list_rules():
    return rating_policies



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import decimal

from ceilometerclient import client as cclient
from keystoneauth1 import loading as ks_loading
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import units

from cloudkitty import collector
from cloudkitty import utils as ck_utils


LOG = logging.getLogger(__name__)

CEILOMETER_COLLECTOR_OPTS = 'ceilometer_collector'
ceilometer_collector_opts = ks_loading.get_auth_common_conf_options()

cfg.CONF.register_opts(ceilometer_collector_opts, CEILOMETER_COLLECTOR_OPTS)
ks_loading.register_session_conf_options(
    cfg.CONF,
    CEILOMETER_COLLECTOR_OPTS)
ks_loading.register_auth_conf_options(
    cfg.CONF,
    CEILOMETER_COLLECTOR_OPTS)
CONF = cfg.CONF

METRICS_CONF = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)


class ResourceNotFound(Exception):
    """Raised when the resource doesn't exist."""

    def __init__(self, resource_type, resource_id):
        super(ResourceNotFound, self).__init__(
            "No such resource: %s, type: %s" % (resource_id, resource_type))
        self.resource_id = resource_id
        self.resource_type = resource_type


class CeilometerResourceCacher(object):
    def __init__(self):
        self._resource_cache = {}

    def add_resource_detail(self, resource_type, resource_id, resource_data):
        if resource_type not in self._resource_cache:
            self._resource_cache[resource_type] = {}
        self._resource_cache[resource_type][resource_id] = resource_data
        return self._resource_cache[resource_type][resource_id]

    def has_resource_detail(self, resource_type, resource_id):
        if resource_type in self._resource_cache:
            if resource_id in self._resource_cache[resource_type]:
                return True
        return False

    def get_resource_detail(self, resource_type, resource_id):
        try:
            resource = self._resource_cache[resource_type][resource_id]
            return resource
        except KeyError:
            raise ResourceNotFound(resource_type, resource_id)


class CeilometerCollector(collector.BaseCollector):
    collector_name = 'ceilometer'
    dependencies = ('CeilometerTransformer',
                    'CloudKittyFormatTransformer')

    units_mappings = {
        'compute': 'instance',
        'image': 'MiB',
        'volume': 'GiB',
        'network.bw.out': 'MB',
        'network.bw.in': 'MB',
        'network.floating': 'ip',
        'radosgw.usage': 'GiB'
    }

    def __init__(self, transformers, **kwargs):
        super(CeilometerCollector, self).__init__(transformers, **kwargs)

        self.t_ceilometer = self.transformers['CeilometerTransformer']
        self.t_cloudkitty = self.transformers['CloudKittyFormatTransformer']

        self._cacher = CeilometerResourceCacher()

        self.auth = ks_loading.load_auth_from_conf_options(
            CONF,
            CEILOMETER_COLLECTOR_OPTS)
        self.session = ks_loading.load_session_from_conf_options(
            CONF,
            CEILOMETER_COLLECTOR_OPTS,
            auth=self.auth)
        self._conn = cclient.get_client(
            '2',
            session=self.session)

    @classmethod
    def get_metadata(cls, resource_name, transformers):
        info = super(CeilometerCollector, cls).get_metadata(resource_name,
                                                            transformers)
        try:
            info["metadata"].extend(transformers['CeilometerTransformer']
                                    .get_metadata(resource_name))

            try:
                tmp = METRICS_CONF['metrics_units'][resource_name]
                info['unit'] = list(tmp.values())[0]['unit']
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated oslo config method.')
                info['unit'] = cls.units_mappings[resource_name]

        except KeyError:
            pass
        return info

    def gen_filter(self, op='eq', **kwargs):
        """Generate ceilometer filter from kwargs."""
        q_filter = []
        for kwarg in kwargs:
            q_filter.append({'field': kwarg, 'op': op, 'value': kwargs[kwarg]})
        return q_filter

    def prepend_filter(self, prepend, **kwargs):
        """Filter composer."""
        q_filter = {}
        for kwarg in kwargs:
            q_filter[prepend + kwarg] = kwargs[kwarg]
        return q_filter

    def user_metadata_filter(self, op='eq', **kwargs):
        """Create user_metadata filter from kwargs."""
        user_filter = {}
        for kwarg in kwargs:
            field = kwarg
            # Auto replace of . to _ to match ceilometer behaviour
            if '.' in field:
                field = field.replace('.', '_')
            user_filter[field] = kwargs[kwarg]
        user_filter = self.prepend_filter('user_metadata.', **user_filter)
        return self.metadata_filter(op, **user_filter)

    def metadata_filter(self, op='eq', **kwargs):
        """Create metadata filter from kwargs."""
        meta_filter = self.prepend_filter('metadata.', **kwargs)
        return self.gen_filter(op, **meta_filter)

    def resources_stats(self,
                        meter,
                        start,
                        end=None,
                        project_id=None,
                        q_filter=None):
        """Resources statistics during the timespan."""
        start_iso = ck_utils.ts2iso(start)
        req_filter = self.gen_filter(op='ge', timestamp=start_iso)
        if project_id:
            req_filter.extend(self.gen_filter(project=project_id))
        if end:
            end_iso = ck_utils.ts2iso(end)
            req_filter.extend(self.gen_filter(op='le', timestamp=end_iso))
        if isinstance(q_filter, list):
            req_filter.extend(q_filter)
        elif q_filter:
            req_filter.append(q_filter)
        resources_stats = self._conn.statistics.list(meter_name=meter,
                                                     period=0, q=req_filter,
                                                     groupby=['resource_id'])
        return resources_stats

    def active_resources(self,
                         meter,
                         start,
                         end=None,
                         project_id=None,
                         q_filter=None):
        """Resources that were active during the timespan."""
        resources_stats = self.resources_stats(meter,
                                               start,
                                               end,
                                               project_id,
                                               q_filter)
        return [resource.groupby['resource_id']
                for resource in resources_stats]

    def get_compute(self, start, end=None, project_id=None, q_filter=None):
        active_instance_ids = self.active_resources('cpu', start, end,
                                                    project_id, q_filter)
        compute_data = []
        for instance_id in active_instance_ids:
            if not self._cacher.has_resource_detail('compute', instance_id):
                raw_resource = self._conn.resources.get(instance_id)
                instance = self.t_ceilometer.strip_resource_data('compute',
                                                                 raw_resource)
                self._cacher.add_resource_detail('compute',
                                                 instance_id,
                                                 instance)
            instance = self._cacher.get_resource_detail('compute',
                                                        instance_id)

            try:
                met = list(METRICS_CONF['metrics_units']['compute'].values())
                compute_data.append(self.t_cloudkitty.format_item(
                    instance,
                    met[0]['unit'],
                    1,
                ))
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated oslo config method.')
                compute_data.append(self.t_cloudkitty.format_item(
                    instance,
                    self.units_mappings['compute'],
                    1,
                ))
        if not compute_data:
            raise collector.NoDataCollected(self.collector_name, 'compute')
        return self.t_cloudkitty.format_service('compute', compute_data)

    def get_image(self, start, end=None, project_id=None, q_filter=None):
        active_image_stats = self.resources_stats('image.size',
                                                  start,
                                                  end,
                                                  project_id,
                                                  q_filter)
        image_data = []
        for image_stats in active_image_stats:
            image_id = image_stats.groupby['resource_id']
            if not self._cacher.has_resource_detail('image', image_id):
                raw_resource = self._conn.resources.get(image_id)
                image = self.t_ceilometer.strip_resource_data('image',
                                                              raw_resource)
                self._cacher.add_resource_detail('image',
                                                 image_id,
                                                 image)
            image = self._cacher.get_resource_detail('image',
                                                     image_id)

            # Unit conversion
            try:
                conv_data = METRICS_CONF['metrics_units']['image']
                image_size_mb = ck_utils.convert_unit(
                    decimal.Decimal(image_stats.max),
                    conv_data['image.size'].get('factor', 1),
                    conv_data['image.size'].get('offset', 0),
                )
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated hardcoded method.')
                image_size_mb = decimal.Decimal(image_stats.max) / units.Mi

            try:
                met = list(METRICS_CONF['metrics_units']['image'].values())
                image_data.append(self.t_cloudkitty.format_item(
                    image,
                    met[0]['unit'],
                    image_size_mb,
                ))
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated oslo config method.')
                image_data.append(self.t_cloudkitty.format_item(
                    image,
                    self.units_mappings['image'],
                    image_size_mb,
                ))

        if not image_data:
            raise collector.NoDataCollected(self.collector_name, 'image')
        return self.t_cloudkitty.format_service('image', image_data)

    def get_volume(self, start, end=None, project_id=None, q_filter=None):
        active_volume_stats = self.resources_stats('volume.size',
                                                   start,
                                                   end,
                                                   project_id,
                                                   q_filter)
        volume_data = []
        for volume_stats in active_volume_stats:
            volume_id = volume_stats.groupby['resource_id']
            if not self._cacher.has_resource_detail('volume',
                                                    volume_id):
                raw_resource = self._conn.resources.get(volume_id)
                volume = self.t_ceilometer.strip_resource_data('volume',
                                                               raw_resource)
                self._cacher.add_resource_detail('volume',
                                                 volume_id,
                                                 volume)
            volume = self._cacher.get_resource_detail('volume',
                                                      volume_id)

            # Unit conversion
            try:
                conv_data = METRICS_CONF['metrics_units']['volume']
                volume_stats.max = ck_utils.convert_unit(
                    decimal.Decimal(volume_stats.max),
                    conv_data['volume.size'].get('factor', 1),
                    conv_data['volume.size'].get('offset', 0),
                )

                volume_data.append(self.t_cloudkitty.format_item(
                    volume,
                    conv_data['volume.size']['unit'],
                    volume_stats.max,
                ))
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated oslo config method.')
                volume_data.append(self.t_cloudkitty.format_item(
                    volume,
                    self.units_mappings['volume'],
                    volume_stats.max,
                ))

        if not volume_data:
            raise collector.NoDataCollected(self.collector_name, 'volume')
        return self.t_cloudkitty.format_service('volume', volume_data)

    def _get_network_bw(self,
                        direction,
                        start,
                        end=None,
                        project_id=None,
                        q_filter=None):
        if direction == 'in':
            resource_type = 'network.incoming.bytes'
        else:
            direction = 'out'
            resource_type = 'network.outgoing.bytes'
        active_tap_stats = self.resources_stats(resource_type,
                                                start,
                                                end,
                                                project_id,
                                                q_filter)
        bw_data = []
        for tap_stat in active_tap_stats:
            tap_id = tap_stat.groupby['resource_id']
            if not self._cacher.has_resource_detail('network.tap',
                                                    tap_id):
                raw_resource = self._conn.resources.get(tap_id)
                tap = self.t_ceilometer.strip_resource_data(
                    'network.tap',
                    raw_resource)
                self._cacher.add_resource_detail('network.tap',
                                                 tap_id,
                                                 tap)
            tap = self._cacher.get_resource_detail('network.tap',
                                                   tap_id)

            # Unit conversion
            try:
                conv = METRICS_CONF['metrics_units']['network.bw.' + direction]
                tap_bw_mb = ck_utils.convert_unit(
                    decimal.Decimal(tap_stat.max),
                    conv[resource_type].get('factor', 1),
                    conv[resource_type].get('offset', 0),
                )
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated hardcoded method.')
                tap_bw_mb = decimal.Decimal(tap_stat.max) / units.M

            try:
                met = METRICS_CONF['metrics_units']['network.bw.' + direction]
                bw_data.append(self.t_cloudkitty.format_item(
                    tap,
                    list(met.values())[0]['unit'],
                    tap_bw_mb,
                ))
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated oslo config method.')
                bw_data.append(self.t_cloudkitty.format_item(
                    tap,
                    self.units_mappings['network.bw.' + direction],
                    tap_bw_mb,
                ))

        ck_res_name = 'network.bw.{}'.format(direction)
        if not bw_data:
            raise collector.NoDataCollected(self.collector_name,
                                            ck_res_name)
        return self.t_cloudkitty.format_service(ck_res_name,
                                                bw_data)

    def get_network_bw_out(self,
                           start,
                           end=None,
                           project_id=None,
                           q_filter=None):
        return self._get_network_bw('out', start, end, project_id, q_filter)

    def get_network_bw_in(self,
                          start,
                          end=None,
                          project_id=None,
                          q_filter=None):
        return self._get_network_bw('in', start, end, project_id, q_filter)

    def get_network_floating(self,
                             start,
                             end=None,
                             project_id=None,
                             q_filter=None):
        active_floating_ids = self.active_resources('ip.floating',
                                                    start,
                                                    end,
                                                    project_id,
                                                    q_filter)
        floating_data = []
        for floating_id in active_floating_ids:
            if not self._cacher.has_resource_detail('network.floating',
                                                    floating_id):
                raw_resource = self._conn.resources.get(floating_id)
                floating = self.t_ceilometer.strip_resource_data(
                    'network.floating',
                    raw_resource)
                self._cacher.add_resource_detail('network.floating',
                                                 floating_id,
                                                 floating)
            floating = self._cacher.get_resource_detail('network.floating',
                                                        floating_id)

            try:
                metric = METRICS_CONF['metrics_units']['network.floating']
                floating_data.append(self.t_cloudkitty.format_item(
                    floating,
                    list(metric.values())[0]['unit'],
                    1,
                ))
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated oslo config method.')
                floating_data.append(self.t_cloudkitty.format_item(
                    floating,
                    self.units_mappings['network.floating'],
                    1,
                ))

        if not floating_data:
            raise collector.NoDataCollected(self.collector_name,
                                            'network.floating')
        return self.t_cloudkitty.format_service('network.floating',
                                                floating_data)

    def get_radosgw_usage(self,
                          start,
                          end=None,
                          project_id=None,
                          q_filter=None):
        active_rgw_stats = self.resources_stats('radosgw.objects.size', start,
                                                end, project_id, q_filter)
        rgw_data = []
        for rgw_stats in active_rgw_stats:
            rgw_id = rgw_stats.groupby['resource_id']
            if not self._cacher.has_resource_detail('radosgw.usage', rgw_id):
                raw_resource = self._conn.resources.get(rgw_id)
                rgw = self.t_ceilometer.strip_resource_data('radosgw.usage',
                                                            raw_resource)
                self._cacher.add_resource_detail('radosgw.usage', rgw_id, rgw)
            rgw = self._cacher.get_resource_detail('radosgw.usage', rgw_id)

            # Unit conversion
            try:
                conv_data = METRICS_CONF['metrics_units']['radosgw.usage']
                rgw_size = ck_utils.convert_unit(
                    decimal.Decimal(rgw_stats.max),
                    conv_data['radosgw.object.size'].get('factor', 1),
                    conv_data['radosgw.object.size'].get('offset', 0),
                )

                rgw_data.append(
                    self.t_cloudkitty.format_item(
                        rgw,
                        conv_data['rados.objects.size']['unit'],
                        rgw_size,
                    )
                )
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated hardcoded method.')
                rgw_size = decimal.Decimal(rgw_stats.max) / units.Gi
                rgw_data.append(
                    self.t_cloudkitty.format_item(
                        rgw,
                        self.units_mappings['radosgw.usage'],
                        rgw_size,
                    )
                )

        if not rgw_data:
            raise collector.NoDataCollected(self.collector_name,
                                            'radosgw.usage')
        return self.t_cloudkitty.format_service('radosgw.usage', rgw_data)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import os

from cloudkitty.common.db.alembic import migration

ALEMBIC_REPO = os.path.join(os.path.dirname(__file__), 'alembic')


def upgrade(revision):
    config = migration.load_alembic_config(ALEMBIC_REPO)
    return migration.upgrade(config, revision)


def version():
    config = migration.load_alembic_config(ALEMBIC_REPO)
    return migration.version(config)


def revision(message, autogenerate):
    config = migration.load_alembic_config(ALEMBIC_REPO)
    return migration.revision(config, message, autogenerate)


def stamp(revision):
    config = migration.load_alembic_config(ALEMBIC_REPO)
    return migration.stamp(config, revision)



# -*- coding: utf-8 -*-
# Copyright 2016 (c) Openstack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Sergio Colinas
#
import datetime
import decimal
import json

import dateutil.parser
from gnocchiclient import client as gclient
from gnocchiclient import exceptions as gexceptions
from keystoneauth1 import loading as ks_loading
from oslo_config import cfg
from oslo_log import log
from oslo_utils import uuidutils
import six

from cloudkitty import storage
from cloudkitty import utils as ck_utils

LOG = log.getLogger(__name__)
CONF = cfg.CONF

METRICS_CONF = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)

GNOCCHI_STORAGE_OPTS = 'storage_gnocchi'
gnocchi_storage_opts = [
    cfg.StrOpt('archive_policy_name',
               default='rating',
               help='Gnocchi storage archive policy name.'),
    # The archive policy definition MUST include the collect period granularity
    cfg.StrOpt('archive_policy_definition',
               default='[{"granularity": '
                       + six.text_type(METRICS_CONF['period']) +
                       ', "timespan": "90 days"}, '
                       '{"granularity": 86400, "timespan": "360 days"}, '
                       '{"granularity": 2592000, "timespan": "1800 days"}]',
               help='Gnocchi storage archive policy definition.'), ]
CONF.register_opts(gnocchi_storage_opts, GNOCCHI_STORAGE_OPTS)

ks_loading.register_session_conf_options(
    CONF,
    GNOCCHI_STORAGE_OPTS)
ks_loading.register_auth_conf_options(
    CONF,
    GNOCCHI_STORAGE_OPTS)

CLOUDKITTY_STATE_RESOURCE = 'cloudkitty_state'
CLOUDKITTY_STATE_METRIC = 'state'


class GnocchiStorage(storage.BaseStorage):
    """Gnocchi Storage Backend.

    Driver used to add full native support for gnocchi, improving performance
    and taking advantage of gnocchi capabilities.
    """

    def __init__(self, **kwargs):
        super(GnocchiStorage, self).__init__(**kwargs)
        self.auth = ks_loading.load_auth_from_conf_options(
            CONF,
            GNOCCHI_STORAGE_OPTS)
        self.session = ks_loading.load_session_from_conf_options(
            CONF,
            GNOCCHI_STORAGE_OPTS,
            auth=self.auth)
        self._conn = gclient.Client('1', session=self.session)
        self._measures = {}
        self._archive_policy_name = (
            CONF.storage_gnocchi.archive_policy_name)
        self._archive_policy_definition = json.loads(
            CONF.storage_gnocchi.archive_policy_definition)
        self._period = METRICS_CONF['period']
        if "period" in kwargs:
            self._period = kwargs["period"]

    def init(self):
        # Creates rating archive-policy if not exists
        try:
            self._conn.archive_policy.get(self._archive_policy_name)
        except gexceptions.ArchivePolicyNotFound:
            ck_policy = {}
            ck_policy["name"] = self._archive_policy_name
            ck_policy["back_window"] = 0
            ck_policy["aggregation_methods"] = ["sum", ]
            ck_policy["definition"] = self._archive_policy_definition
            self._conn.archive_policy.create(ck_policy)
        # Creates state resource if it doesn't exist
        try:
            self._conn.resource_type.create(
                {'name': CLOUDKITTY_STATE_RESOURCE})
        except gexceptions.ResourceAlreadyExists:
            pass

    def _get_or_create_resource(self, resource_type, tenant_id):
        """Return the id of a resource or create it.

        :param resource_type: The type of the resource.
        :type metric_name: str
        :param tenant_id: Owner's resource tenant id.
        :type metric_name: str
        """
        query = {"=": {"project_id": tenant_id}}
        resources = self._conn.resource.search(
            resource_type=resource_type,
            query=query,
            limit=1)
        if not resources:
            # NOTE(sheeprine): We don't have the user id information and we are
            # doing rating on a per tenant basis. Put garbage in it
            resource = self._conn.resource.create(
                resource_type=resource_type,
                resource={'id': uuidutils.generate_uuid(),
                          'user_id': None,
                          'project_id': tenant_id})
            return resource['id']
        return resources[0]['id']

    def _get_or_create_metric(self, metric_name, resource_id):
        """Return the metric id from a metric or create it.

        :param metric_name: The name of the metric.
        :type metric_name: str
        :param resource_id: Resource id containing the metric.
        :type metric_name: str
        """
        resource = self._conn.resource.get(
            resource_type='generic',
            resource_id=resource_id,
            history=False)
        metric_id = resource["metrics"].get(metric_name)
        if not metric_id:
            new_metric = {}
            new_metric["archive_policy_name"] = self._archive_policy_name
            new_metric["name"] = metric_name
            new_metric["resource_id"] = resource_id
            metric = self._conn.metric.create(new_metric)
            metric_id = metric['id']
        return metric_id

    def _pre_commit(self, tenant_id):
        measures = self._measures.pop(tenant_id, {})
        self._measures[tenant_id] = dict()
        for resource_id, metrics in measures.items():
            total = metrics.pop('total.cost')
            total_id = self._get_or_create_metric(
                'total.cost',
                resource_id)
            # TODO(sheeprine): Find a better way to handle total
            total_value = sum([decimal.Decimal(val["value"]) for val in total])
            total_timestamp = max([dateutil.parser.parse(val["timestamp"])
                                   for val in total])
            self._measures[tenant_id][total_id] = [{
                'timestamp': total_timestamp.isoformat(),
                'value': six.text_type(total_value)}]
            for metric_name, values in metrics.items():
                metric_id = self._get_or_create_metric(
                    metric_name,
                    resource_id)
                self._measures[tenant_id][metric_id] = values
        state_resource_id = self._get_or_create_resource(
            CLOUDKITTY_STATE_RESOURCE,
            tenant_id)
        state_metric_id = self._get_or_create_metric(
            CLOUDKITTY_STATE_METRIC,
            state_resource_id)
        self._measures[tenant_id][state_metric_id] = [{
            'timestamp': self.usage_start_dt.get(tenant_id).isoformat(),
            'value': 1}]

    def _commit(self, tenant_id):
        if tenant_id in self._measures:
            self._conn.metric.batch_metrics_measures(
                self._measures[tenant_id])

    def _post_commit(self, tenant_id):
        super(GnocchiStorage, self)._post_commit(tenant_id)
        if tenant_id in self._measures:
            del self._measures[tenant_id]

    def _append_metric(self, resource_id, metric_name, value, tenant_id):
        sample = {}
        sample["timestamp"] = self.usage_start_dt.get(tenant_id).isoformat()
        sample["value"] = six.text_type(value)
        measures = self._measures.get(tenant_id) or dict()
        if not measures:
            self._measures[tenant_id] = measures
        metrics = measures.get(resource_id) or dict()
        if not metrics:
            measures[resource_id] = metrics
        metrics[metric_name] = [sample]

    def _dispatch(self, data, tenant_id):
        for metric_name, metrics in data.items():
            for item in metrics:
                resource_id = item["desc"]["resource_id"]
                price = item["rating"]["price"]
                self._append_metric(
                    resource_id,
                    metric_name + ".cost",
                    price,
                    tenant_id)
                self._append_metric(
                    resource_id,
                    'total.cost',
                    price,
                    tenant_id)
                self._has_data[tenant_id] = True

    def set_state(self, state, tenant_id):
        state_resource_id = self._get_or_create_resource(
            CLOUDKITTY_STATE_RESOURCE,
            tenant_id)
        state_metric_id = self._get_or_create_metric(
            CLOUDKITTY_STATE_METRIC,
            state_resource_id)
        self._conn.metric.add_measures(
            state_metric_id,
            [{'timestamp': state.isoformat(),
             'value': 1}])

    def get_state(self, tenant_id=None):
        # Return the last written frame's timestamp.
        query = {"=": {"project_id": tenant_id}} if tenant_id else {}
        state_resource_id = self._get_or_create_resource(
            CLOUDKITTY_STATE_RESOURCE,
            tenant_id)
        try:
            # (aolwas) add "refresh=True" to be sure to get all posted
            # measures for this particular metric
            r = self._conn.metric.get_measures(
                metric=CLOUDKITTY_STATE_METRIC,
                resource_id=state_resource_id,
                query=query,
                aggregation="sum",
                limit=1,
                granularity=self._period,
                needed_overlap=0,
                refresh=True)
        except gexceptions.MetricNotFound:
            return
        if len(r) > 0:
            # NOTE(lukapeschke) Since version 5.0.0, gnocchiclient returns a
            # datetime object instead of a timestamp. This fixture is made
            # to ensure compatibility with all versions
            try:
                # (aolwas) According http://gnocchi.xyz/rest.html#metrics,
                # gnocchi always returns measures ordered by timestamp
                return ck_utils.dt2ts(dateutil.parser.parse(r[-1][0]))
            except TypeError:
                return ck_utils.dt2ts(r[-1][0])

    def get_total(self, begin=None, end=None, tenant_id=None,
                  service=None, groupby=None):
        # Get total rate in timeframe from gnocchi
        metric = "total.cost"
        if service:
            metric = service + ".cost"
        # We need to pass a query to force a post in gnocchi client metric
        # aggregation, so we use one that always meets
        query = {"and": [{">": {"started_at": "1900-01-01T00:00"}}]}
        if tenant_id:
            query = {"=": {"project_id": tenant_id}}
        # TODO(Aaron): need support with groupby
        if groupby:
            LOG.warning('Now get total with groupby not support '
                        'in gnocchi storage backend')
        # TODO(sheeprine): Use server side aggregation
        r = self._conn.metric.aggregation(metrics=metric, query=query,
                                          start=begin, stop=end,
                                          aggregation="sum",
                                          granularity=self._period,
                                          needed_overlap=0)

        rate = sum([measure[2] for measure in r]) if len(r) else 0
        # Return a list of dict
        totallist = []
        total = dict(begin=begin, end=end, rate=rate)
        totallist.append(total)
        return totallist

    def get_tenants(self, begin, end):
        # We need to pass a query to force a post in gnocchi client metric
        # aggregation, so we use one that always meets
        query = {'=': {'type': 'cloudkitty_state'}}
        r = self._conn.metric.aggregation(
            metrics=CLOUDKITTY_STATE_METRIC,
            query=query,
            start=begin,
            stop=end,
            aggregation="sum",
            granularity=self._period,
            needed_overlap=0,
            resource_type=CLOUDKITTY_STATE_RESOURCE,
            groupby="project_id")
        projects = [measures["group"]["project_id"]
                    for measures in r if len(measures["measures"])]
        if len(projects) > 0:
            return projects
        return []

    def _get_resource_data(self, res_type, resource_id, begin, end):
        # Get resource information from gnocchi
        return self._collector.resource_info(
            resource_name=res_type,
            start=begin,
            end=end,
            project_id=None)

    def _to_cloudkitty(self, res_type, resource_data, measure):
        # NOTE(lukapeschke) Since version 5.0.0, gnocchiclient returns a
        # datetime object instead of a timestamp. This fixture is made
        # to ensure compatibility with all versions
        try:
            begin = dateutil.parser.parse(measure[0])
            end = (dateutil.parser.parse(measure[0]) +
                   datetime.timedelta(seconds=self._period))
        except TypeError:
            begin = measure[0]
            end = begin + datetime.timedelta(seconds=self._period)
        cost = decimal.Decimal(measure[2])
        # Rating informations
        rating_dict = {}
        rating_dict['price'] = cost

        # Encapsulate informations in a resource dict
        res_dict = {}
        # TODO(sheeprine): Properly recurse on elements
        resource_data = resource_data[0]
        res_dict['desc'] = resource_data['desc']
        if "qty" in resource_data["vol"]:
            resource_data["vol"]["qty"] = (
                decimal.Decimal(resource_data["vol"]["qty"]))
        res_dict['vol'] = resource_data['vol']
        res_dict['rating'] = rating_dict
        res_dict['tenant_id'] = resource_data['desc']['project_id']

        # Add resource to the usage dict
        usage_dict = {}
        usage_dict[res_type] = [res_dict]

        # Time informations
        period_dict = {}
        period_dict['begin'] = begin.isoformat()
        period_dict['end'] = end.isoformat()

        # Add period to the resource informations
        ck_dict = {}
        ck_dict['period'] = period_dict
        ck_dict['usage'] = usage_dict
        return ck_dict

    def get_time_frame(self, begin, end, **filters):
        tenant_id = filters.get('tenant_id')
        query = dict()
        if tenant_id:
            query['='] = {'project_id': tenant_id}
        else:
            # NOTE(sheeprine): Dummy filter to comply with gnocchi
            query['!='] = {'project_id': None}
        try:
            res_map = METRICS_CONF['services_objects']
        except KeyError:
            res_map = self._collector.retrieve_mappings
        res_type = filters.get('res_type')
        resources = [res_type] if res_type else res_map.keys()
        ck_res = []
        for resource in resources:
            resource_type = res_map[resource]
            r = self._conn.metric.aggregation(
                metrics=resource + ".cost",
                resource_type=resource_type,
                query=query,
                start=begin,
                stop=end,
                granularity=self._period,
                aggregation="sum",
                needed_overlap=0,
                groupby=["type", "id"])
            for resource_measures in r:
                resource_data = None
                for measure in resource_measures["measures"]:
                    if not resource_data:
                        resource_data = self._get_resource_data(
                            res_type=resource,
                            resource_id=resource_measures["group"]["id"],
                            begin=begin,
                            end=end)
                    ck_res.append(
                        self._to_cloudkitty(
                            res_type=resource,
                            resource_data=resource_data,
                            measure=measure))
        return ck_res



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import decimal

from cloudkitty import rating
from cloudkitty.rating.hash.controllers import root as root_api
from cloudkitty.rating.hash.db import api as hash_db_api


class HashMap(rating.RatingProcessorBase):
    """HashMap rating module.

    HashMap can be used to map arbitrary fields of a resource to different
    costs.
    """

    module_name = 'hashmap'
    description = 'HashMap rating module.'
    hot_config = True
    config_controller = root_api.HashMapConfigController

    db_api = hash_db_api.get_instance()

    def __init__(self, tenant_id=None):
        super(HashMap, self).__init__(tenant_id)
        self._entries = {}
        self._res = {}
        self._load_rates()

    def reload_config(self):
        """Reload the module's configuration.

        """
        self._load_rates()

    def _load_mappings(self, mappings_uuid_list):
        hashmap = hash_db_api.get_instance()
        mappings = {}
        for mapping_uuid in mappings_uuid_list:
            mapping_db = hashmap.get_mapping(uuid=mapping_uuid)
            if mapping_db.group_id:
                group_name = mapping_db.group.name
            else:
                group_name = '_DEFAULT_'
            if group_name not in mappings:
                mappings[group_name] = {}
            current_scope = mappings[group_name]

            mapping_value = mapping_db.value
            if mapping_value:
                current_scope[mapping_value] = {}
                current_scope = current_scope[mapping_value]
            current_scope['type'] = mapping_db.map_type
            current_scope['cost'] = mapping_db.cost
        return mappings

    def _load_thresholds(self, thresholds_uuid_list):
        hashmap = hash_db_api.get_instance()
        thresholds = {}
        for threshold_uuid in thresholds_uuid_list:
            threshold_db = hashmap.get_threshold(uuid=threshold_uuid)
            if threshold_db.group_id:
                group_name = threshold_db.group.name
            else:
                group_name = '_DEFAULT_'
            if group_name not in thresholds:
                thresholds[group_name] = {}
            current_scope = thresholds[group_name]

            threshold_level = threshold_db.level
            current_scope[threshold_level] = {}
            current_scope = current_scope[threshold_level]
            current_scope['type'] = threshold_db.map_type
            current_scope['cost'] = threshold_db.cost
        return thresholds

    def _update_entries(self,
                        entry_type,
                        root,
                        service_uuid=None,
                        field_uuid=None,
                        tenant_uuid=None):
        hashmap = hash_db_api.get_instance()
        list_func = getattr(hashmap, 'list_{}'.format(entry_type))
        entries_uuid_list = list_func(
            service_uuid=service_uuid,
            field_uuid=field_uuid,
            tenant_uuid=tenant_uuid)
        load_func = getattr(self, '_load_{}'.format(entry_type))
        entries = load_func(entries_uuid_list)
        if entry_type in root:
            res = root[entry_type]
            for group, values in entries.items():
                if group in res:
                    res[group].update(values)
                else:
                    res[group] = values
        else:
            root[entry_type] = entries

    def _load_service_entries(self, service_name, service_uuid):
        self._entries[service_name] = dict()
        for entry_type in ('mappings', 'thresholds'):
            for tenant in (None, self._tenant_id):
                self._update_entries(
                    entry_type,
                    self._entries[service_name],
                    service_uuid=service_uuid,
                    tenant_uuid=tenant)

    def _load_field_entries(self, service_name, field_name, field_uuid):
        if service_name not in self._entries:
            self._entries[service_name] = {}
        if 'fields' not in self._entries[service_name]:
            self._entries[service_name]['fields'] = {}
        scope = self._entries[service_name]['fields'][field_name] = {}
        for entry_type in ('mappings', 'thresholds'):
            for tenant in (None, self._tenant_id):
                self._update_entries(
                    entry_type,
                    scope,
                    field_uuid=field_uuid,
                    tenant_uuid=tenant)

    def _load_rates(self):
        self._entries = {}
        hashmap = hash_db_api.get_instance()
        services_uuid_list = hashmap.list_services()
        for service_uuid in services_uuid_list:
            service_db = hashmap.get_service(uuid=service_uuid)
            service_name = service_db.name
            self._load_service_entries(service_name, service_uuid)
            fields_uuid_list = hashmap.list_fields(service_uuid)
            for field_uuid in fields_uuid_list:
                field_db = hashmap.get_field(uuid=field_uuid)
                field_name = field_db.name
                self._load_field_entries(service_name, field_name, field_uuid)

    def add_rating_informations(self, data):
        if 'rating' not in data:
            data['rating'] = {'price': 0}
        for entry in self._res.values():
            rate = entry['rate']
            flat = entry['flat']
            if entry['threshold']['scope'] == 'field':
                if entry['threshold']['type'] == 'flat':
                    flat += entry['threshold']['cost']
                else:
                    rate *= entry['threshold']['cost']
            res = rate * flat
            # FIXME(sheeprine): Added here to ensure that qty is decimal
            res *= decimal.Decimal(data['vol']['qty'])
            if entry['threshold']['scope'] == 'service':
                if entry['threshold']['type'] == 'flat':
                    res += entry['threshold']['cost']
                else:
                    res *= entry['threshold']['cost']
            data['rating']['price'] += res

    def update_result(self,
                      group,
                      map_type,
                      cost,
                      level=0,
                      is_threshold=False,
                      threshold_scope='field'):
        if group not in self._res:
            self._res[group] = {'flat': 0,
                                'rate': 1,
                                'threshold': {
                                    'level': -1,
                                    'cost': 0,
                                    'type': 'flat',
                                    'scope': 'field'}}
        if is_threshold:
            best = self._res[group]['threshold']['level']
            if level > best:
                self._res[group]['threshold']['level'] = level
                self._res[group]['threshold']['cost'] = cost
                self._res[group]['threshold']['type'] = map_type
                self._res[group]['threshold']['scope'] = threshold_scope
        else:
            if map_type == 'rate':
                self._res[group]['rate'] *= cost
            elif map_type == 'flat':
                new_flat = cost
                cur_flat = self._res[group]['flat']
                if new_flat > cur_flat:
                    self._res[group]['flat'] = new_flat

    def process_mappings(self,
                         mapping_groups,
                         cmp_value):
        for group_name, mappings in mapping_groups.items():
            for mapping_value, mapping in mappings.items():
                if cmp_value == mapping_value:
                    self.update_result(
                        group_name,
                        mapping['type'],
                        mapping['cost'])

    def process_thresholds(self,
                           threshold_groups,
                           cmp_level,
                           threshold_type):
        for group_name, thresholds in threshold_groups.items():
            for threshold_level, threshold in thresholds.items():
                if cmp_level >= threshold_level:
                    self.update_result(
                        group_name,
                        threshold['type'],
                        threshold['cost'],
                        threshold_level,
                        True,
                        threshold_type)

    def process_services(self, service_name, data):
        if service_name not in self._entries:
            return
        service_mappings = self._entries[service_name]['mappings']
        for group_name, mapping in service_mappings.items():
            self.update_result(group_name,
                               mapping['type'],
                               mapping['cost'])
        service_thresholds = self._entries[service_name]['thresholds']
        self.process_thresholds(service_thresholds,
                                data['vol']['qty'],
                                'service')

    def process_fields(self, service_name, data):
        if service_name not in self._entries:
            return
        if 'fields' not in self._entries[service_name]:
            return
        desc_data = data['desc']
        field_mappings = self._entries[service_name]['fields']
        for field_name, group_mappings in field_mappings.items():
            if field_name not in desc_data:
                continue
            cmp_value = desc_data[field_name]
            self.process_mappings(group_mappings['mappings'],
                                  cmp_value)
            if group_mappings['thresholds']:
                self.process_thresholds(group_mappings['thresholds'],
                                        decimal.Decimal(cmp_value),
                                        'field')

    def process(self, data):
        for cur_data in data:
            cur_usage = cur_data['usage']
            for service_name, service_data in cur_usage.items():
                for item in service_data:
                    self._res = {}
                    self.process_services(service_name, item)
                    self.process_fields(service_name, item)
                    self.add_rating_informations(item)
        return data



# Copyright 2017 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""improve_qty_digit

Revision ID: c703a1bad612
Revises: 307430ab38bc
Create Date: 2017-04-01 09:33:41.434750

"""

# revision identifiers, used by Alembic.
revision = 'c703a1bad612'
down_revision = '307430ab38bc'

from alembic import op
import sqlalchemy as sa


def upgrade():
    with op.batch_alter_table('rated_data_frames') as batch_op:
        batch_op.alter_column(
            'qty',
            type_=sa.Numeric(15, 5),
            existing_type=sa.Numeric())



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""improve qty precision

Revision ID: 307430ab38bc
Revises: 792b438b663
Create Date: 2016-09-05 18:37:26.714065

"""

# revision identifiers, used by Alembic.
revision = '307430ab38bc'
down_revision = '792b438b663'

from alembic import op
import sqlalchemy as sa


def upgrade():
    with op.batch_alter_table('rated_data_frames') as batch_op:
        batch_op.alter_column(
            'qty',
            type_=sa.Numeric(10, 5),
            existing_type=sa.Numeric())



# -*- coding: utf-8 -*-
# Copyright 2017 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Luka Peschke
#
"""initial revision

Revision ID: 03da4bb002b9
Revises: None
Create Date: 2017-11-21 15:59:26.776639

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '03da4bb002b9'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'hybrid_storage_states',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('tenant_id', sa.String(length=32), nullable=False),
        sa.Column('state', sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_charset='utf8',
        mysql_engine='InnoDB')



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import decimal
import json

from oslo_db.sqlalchemy import utils
import sqlalchemy

from cloudkitty import db
from cloudkitty import storage
from cloudkitty.storage.sqlalchemy import migration
from cloudkitty.storage.sqlalchemy import models
from cloudkitty import utils as ck_utils


class DecimalJSONEncoder(json.JSONEncoder):
    """Wrapper class to handle decimal.Decimal objects in json.dumps()."""
    def default(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        return super(DecimalJSONEncoder, self).default(obj)


class SQLAlchemyStorage(storage.BaseStorage):
    """SQLAlchemy Storage Backend

    """
    frame_model = models.RatedDataFrame

    def __init__(self, **kwargs):
        super(SQLAlchemyStorage, self).__init__(**kwargs)
        self._session = {}

    @staticmethod
    def init():
        migration.upgrade('head')

    def _pre_commit(self, tenant_id):
        self._check_session(tenant_id)
        if not self._has_data.get(tenant_id):
            empty_frame = {'vol': {'qty': 0, 'unit': 'None'},
                           'rating': {'price': 0}, 'desc': ''}
            self._append_time_frame('_NO_DATA_', empty_frame, tenant_id)

    def _commit(self, tenant_id):
        self._session[tenant_id].commit()

    def _post_commit(self, tenant_id):
        super(SQLAlchemyStorage, self)._post_commit(tenant_id)
        del self._session[tenant_id]

    def _check_session(self, tenant_id):
        session = self._session.get(tenant_id)
        if not session:
            self._session[tenant_id] = db.get_session()
            self._session[tenant_id].begin()

    def _dispatch(self, data, tenant_id):
        self._check_session(tenant_id)
        for service in data:
            for frame in data[service]:
                self._append_time_frame(service, frame, tenant_id)
                self._has_data[tenant_id] = True

    def get_state(self, tenant_id=None):
        session = db.get_session()
        q = utils.model_query(
            self.frame_model,
            session)
        if tenant_id:
            q = q.filter(
                self.frame_model.tenant_id == tenant_id)
        q = q.order_by(
            self.frame_model.begin.desc())
        r = q.first()
        if r:
            return ck_utils.dt2ts(r.begin)

    def get_total(self, begin=None, end=None, tenant_id=None,
                  service=None, groupby=None):
        session = db.get_session()
        querymodels = [
            sqlalchemy.func.sum(self.frame_model.rate).label('rate')
        ]

        # Boundary calculation
        if not begin:
            begin = ck_utils.get_month_start()
        if not end:
            end = ck_utils.get_next_month()
        if tenant_id:
            querymodels.append(self.frame_model.tenant_id)
        if service:
            querymodels.append(self.frame_model.res_type)
        if groupby:
            groupbyfields = groupby.split(",")
            for field in groupbyfields:
                field_obj = self.frame_model.__dict__.get(field, None)
                if field_obj and field_obj not in querymodels:
                    querymodels.append(field_obj)

        q = session.query(*querymodels)
        if tenant_id:
            q = q.filter(
                self.frame_model.tenant_id == tenant_id)
        if service:
            q = q.filter(
                self.frame_model.res_type == service)
        q = q.filter(
            self.frame_model.begin >= begin,
            self.frame_model.end <= end,
            self.frame_model.res_type != '_NO_DATA_')
        if groupby:
            q = q.group_by(groupby)

        # Order by sum(rate)
        q = q.order_by(sqlalchemy.func.sum(self.frame_model.rate))
        results = q.all()
        totallist = []
        for r in results:
            total = {model.name: value for model, value in zip(querymodels, r)}
            total["begin"] = begin
            total["end"] = end
            totallist.append(total)
        return totallist

    def get_tenants(self, begin, end):
        session = db.get_session()
        q = utils.model_query(
            self.frame_model,
            session)
        q = q.filter(
            self.frame_model.begin >= begin,
            self.frame_model.end <= end)
        tenants = q.distinct().values(
            self.frame_model.tenant_id)
        return [tenant.tenant_id for tenant in tenants]

    def get_time_frame(self, begin, end, **filters):
        session = db.get_session()
        q = utils.model_query(
            self.frame_model,
            session)
        q = q.filter(
            self.frame_model.begin >= ck_utils.ts2dt(begin),
            self.frame_model.end <= ck_utils.ts2dt(end))
        for filter_name, filter_value in filters.items():
            if filter_value:
                q = q.filter(
                    getattr(self.frame_model, filter_name) == filter_value)
        if not filters.get('res_type'):
            q = q.filter(self.frame_model.res_type != '_NO_DATA_')
        count = q.count()
        if not count:
            raise storage.NoTimeFrame()
        r = q.all()
        return [entry.to_cloudkitty(self._collector) for entry in r]

    def _append_time_frame(self, res_type, frame, tenant_id):
        vol_dict = frame['vol']
        qty = vol_dict['qty']
        unit = vol_dict['unit']
        rating_dict = frame.get('rating', {})
        rate = rating_dict.get('price')
        if not rate:
            rate = decimal.Decimal(0)
        desc = json.dumps(frame['desc'], cls=DecimalJSONEncoder)
        self.add_time_frame(begin=self.usage_start_dt.get(tenant_id),
                            end=self.usage_end_dt.get(tenant_id),
                            tenant_id=tenant_id,
                            unit=unit,
                            qty=qty,
                            res_type=res_type,
                            rate=rate,
                            desc=desc)

    def add_time_frame(self, **kwargs):
        """Create a new time frame.

        :param begin: Start of the dataframe.
        :param end: End of the dataframe.
        :param tenant_id: tenant_id of the dataframe owner.
        :param unit: Unit of the metric.
        :param qty: Quantity of the metric.
        :param res_type: Type of the resource.
        :param rate: Calculated rate for this dataframe.
        :param desc: Resource description (metadata).
        """
        frame = self.frame_model(**kwargs)
        self._session[kwargs.get('tenant_id')].add(frame)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import copy
import itertools

import cloudkitty.api.app
import cloudkitty.collector.ceilometer
import cloudkitty.collector.gnocchi
import cloudkitty.collector.monasca
import cloudkitty.config
import cloudkitty.orchestrator
import cloudkitty.service
import cloudkitty.storage
import cloudkitty.storage.gnocchi
import cloudkitty.tenant_fetcher
import cloudkitty.tenant_fetcher.keystone
import cloudkitty.utils

__all__ = ['list_opts']

_opts = [
    ('api', list(itertools.chain(
        cloudkitty.api.app.api_opts,))),
    ('collect', list(itertools.chain(
        cloudkitty.utils.collect_opts))),
    ('ceilometer_collector', list(itertools.chain(
        cloudkitty.collector.ceilometer.ceilometer_collector_opts))),
    ('collector_monasca', list(itertools.chain(
        cloudkitty.collector.monasca.collector_monasca_opts))),
    ('gnocchi_collector', list(itertools.chain(
        cloudkitty.collector.gnocchi.gnocchi_collector_opts))),
    ('keystone_fetcher', list(itertools.chain(
        cloudkitty.tenant_fetcher.keystone.keystone_fetcher_opts,
        cloudkitty.tenant_fetcher.keystone.keystone_common_opts))),
    ('orchestrator', list(itertools.chain(
        cloudkitty.orchestrator.orchestrator_opts))),
    ('output', list(itertools.chain(
        cloudkitty.config.output_opts))),
    ('state', list(itertools.chain(
        cloudkitty.config.state_opts))),
    ('storage', list(itertools.chain(
        cloudkitty.storage.storage_opts))),
    ('storage_gnocchi', list(itertools.chain(
        cloudkitty.storage.gnocchi.gnocchi_storage_opts))),
    ('tenant_fetcher', list(itertools.chain(
        cloudkitty.tenant_fetcher.fetchers_opts))),
    (None, list(itertools.chain(
        cloudkitty.api.app.auth_opts,
        cloudkitty.service.service_opts)))
]


def list_opts():
    return [(g, copy.deepcopy(o)) for g, o in _opts]



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from stevedore import extension

from cloudkitty import collector
from cloudkitty.db import api as db_api

COLLECTORS_NAMESPACE = 'cloudkitty.collector.backends'


class MetaCollector(collector.BaseCollector):
    def __init__(self, transformers, **kwargs):
        super(MetaCollector, self).__init__(transformers, **kwargs)

        self._db = db_api.get_instance().get_service_to_collector_mapping()

        self._collectors = {}
        self._load_collectors()

        self._mappings = {}
        self._load_mappings()

    def _load_mappings(self):
        mappings = self._db.list_services()
        for mapping in mappings:
            db_mapping = self._db.get_mapping(mapping.service)
            self._mappings[db_mapping.service] = db_mapping.collector

    def _check_enabled(self, name):
        enable_state = db_api.get_instance().get_module_info()
        return enable_state.get_state('collector_{}'.format(name))

    def _load_collectors(self):
        self._collectors = {}
        collectors = extension.ExtensionManager(
            COLLECTORS_NAMESPACE,
        )
        collectors_list = collectors.names()
        collectors_list.remove('meta')

        for name in collectors_list:
            if self._check_enabled(name):
                self._collectors[name] = collectors[name].plugin(
                    self.transformers,
                    period=self.period)

    def retrieve(self,
                 resource,
                 start,
                 end=None,
                 project_id=None,
                 q_filter=None):
        collector_list = self._collectors.values()
        # Set designated collector on top of the list
        try:
            collector_name = self._mappings[resource]
            designated_collector = self._collectors[collector_name]
            collector_list.remove(designated_collector)
            collector_list.insert(0, designated_collector)
        except KeyError:
            pass
        for cur_collector in collector_list:
            # Try every collector until we get a result
            try:
                return cur_collector.retrieve(
                    resource,
                    start,
                    end,
                    project_id,
                    q_filter)
            except NotImplementedError:
                pass
        raise NotImplementedError("No collector for resource '%s'." % resource)



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import pecan
import six
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1 import types as ck_types
from cloudkitty import rating
from cloudkitty.rating.pyscripts.datamodels import script as script_models
from cloudkitty.rating.pyscripts.db import api as db_api


class PyScriptsScriptsController(rating.RatingRestControllerBase):
    """Controller responsible of scripts management.

    """

    def normalize_data(self, data):
        """Translate data to binary format if needed.

        :param data: Data to convert to binary type.
        """
        if data == wtypes.Unset:
            return ''
        if not isinstance(data, six.binary_type):
            data = data.encode('utf-8')
        return data

    @wsme_pecan.wsexpose(script_models.ScriptCollection, bool)
    def get_all(self, no_data=False):
        """Get the script list

        :param no_data: Set to True to remove script data from output.
        :return: List of every scripts.
        """
        pyscripts = db_api.get_instance()
        script_list = []
        script_uuid_list = pyscripts.list_scripts()
        for script_uuid in script_uuid_list:
            script_db = pyscripts.get_script(uuid=script_uuid)
            script = script_db.export_model()
            if no_data:
                del script['data']
            script_list.append(script_models.Script(
                **script))
        res = script_models.ScriptCollection(scripts=script_list)
        return res

    @wsme_pecan.wsexpose(script_models.Script, ck_types.UuidType())
    def get_one(self, script_id):
        """Return a script.

        :param script_id: UUID of the script to filter on.
        """
        pyscripts = db_api.get_instance()
        try:
            script_db = pyscripts.get_script(uuid=script_id)
            return script_models.Script(**script_db.export_model())
        except db_api.NoSuchScript as e:
            pecan.abort(404, six.text_type(e))

    @wsme_pecan.wsexpose(script_models.Script,
                         body=script_models.Script,
                         status_code=201)
    def post(self, script_data):
        """Create pyscripts script.

        :param script_data: Informations about the script to create.
        """
        pyscripts = db_api.get_instance()
        try:
            data = self.normalize_data(script_data.data)
            script_db = pyscripts.create_script(script_data.name, data)
            pecan.response.location = pecan.request.path_url
            if pecan.response.location[-1] != '/':
                pecan.response.location += '/'
            pecan.response.location += script_db.script_id
            return script_models.Script(
                **script_db.export_model())
        except db_api.ScriptAlreadyExists as e:
            pecan.abort(409, six.text_type(e))

    @wsme_pecan.wsexpose(script_models.Script,
                         ck_types.UuidType(),
                         body=script_models.Script,
                         status_code=201)
    def put(self, script_id, script_data):
        """Update pyscripts script.

        :param script_id: UUID of the script to update.
        :param script_data: Script data to update.
        """
        pyscripts = db_api.get_instance()
        try:
            data = self.normalize_data(script_data.data)
            script_db = pyscripts.update_script(script_id,
                                                name=script_data.name,
                                                data=data)
            pecan.response.location = pecan.request.path_url
            if pecan.response.location[-1] != '/':
                pecan.response.location += '/'
            pecan.response.location += script_db.script_id
            return script_models.Script(
                **script_db.export_model())
        except db_api.NoSuchScript as e:
            pecan.abort(404, six.text_type(e))

    @wsme_pecan.wsexpose(None, ck_types.UuidType(), status_code=204)
    def delete(self, script_id):
        """Delete the script.

        :param script_id: UUID of the script to delete.
        """
        pyscripts = db_api.get_instance()
        try:
            pyscripts.delete_script(uuid=script_id)
        except db_api.NoSuchScript as e:
            pecan.abort(404, six.text_type(e))



# -*- coding: utf-8 -*-
# Copyright 2016 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Maxime Cottret
#
from oslo_config import cfg
import pecan
from pecan import rest
import six
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1.datamodels import info as info_models
from cloudkitty.api.v1 import types as ck_types
from cloudkitty import collector
from cloudkitty.common import policy
from cloudkitty import utils as ck_utils

CONF = cfg.CONF

METRICS_CONF = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)

METADATA = collector.get_collector_metadata()


class ServiceInfoController(rest.RestController):
    """REST Controller mananging collected services information."""

    @wsme_pecan.wsexpose(info_models.CloudkittyServiceInfoCollection)
    def get_all(self):
        """Get the service list.

        :return: List of every services.
        """
        policy.authorize(pecan.request.context, 'info:list_services_info', {})
        services_info_list = []
        for service, metadata in METADATA.items():
            info = metadata.copy()
            info['service_id'] = service
            services_info_list.append(
                info_models.CloudkittyServiceInfo(**info))
        return info_models.CloudkittyServiceInfoCollection(
            services=services_info_list)

    @wsme_pecan.wsexpose(info_models.CloudkittyServiceInfo, wtypes.text)
    def get_one(self, service_name):
        """Return a service.

        :param service_name: name of the service.
        """
        policy.authorize(pecan.request.context, 'info:get_service_info', {})
        try:
            info = METADATA[service_name].copy()
            info['service_id'] = service_name
            return info_models.CloudkittyServiceInfo(**info)
        except KeyError:
            pecan.abort(404, six.text_type(service_name))


class InfoController(rest.RestController):
    """REST Controller managing Cloudkitty general information."""

    services = ServiceInfoController()

    _custom_actions = {'config': ['GET']}

    @wsme_pecan.wsexpose({
        str: ck_types.MultiType(wtypes.text, int, float, dict, list)
    })
    def config(self):
        """Return current configuration."""
        policy.authorize(pecan.request.context, 'info:get_config', {})
        info = {}
        info["collect"] = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)
        return info



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_config import cfg
from oslo_utils import fileutils
from stevedore import named

from cloudkitty import state
from cloudkitty import storage
from cloudkitty import utils as ck_utils

CONF = cfg.CONF
WRITERS_NAMESPACE = 'cloudkitty.output.writers'


class WriteOrchestrator(object):
    """Write Orchestrator:

        Handle incoming data from the global orchestrator, and store them in an
        intermediary data format before final transformation.
    """
    def __init__(self,
                 backend,
                 tenant_id,
                 storage,
                 basepath=None,
                 period=3600):
        self._backend = backend
        self._tenant_id = tenant_id
        self._storage = storage
        self._basepath = basepath
        if self._basepath:
            fileutils.ensure_tree(self._basepath)
        self._period = period
        self._sm = state.DBStateManager(self._tenant_id,
                                        'writer_status')
        self._write_pipeline = []

        # State vars
        self.usage_start = None
        self.usage_end = None

        # Current total
        self.total = 0

    def init_writing_pipeline(self):
        CONF.import_opt('pipeline', 'cloudkitty.config', 'output')
        output_pipeline = named.NamedExtensionManager(
            WRITERS_NAMESPACE,
            CONF.output.pipeline)
        for writer in output_pipeline:
            self.add_writer(writer.plugin)

    def add_writer(self, writer_class):
        writer = writer_class(self,
                              self._tenant_id,
                              self._backend,
                              self._basepath)
        self._write_pipeline.append(writer)

    def _update_state_manager_data(self):
        self._sm.set_state(self.usage_end)
        metadata = {'total': self.total}
        self._sm.set_metadata(metadata)

    def _load_state_manager_data(self):
        timeframe = self._sm.get_state()
        if timeframe:
            self.usage_start = timeframe
            self.usage_end = self.usage_start + self._period
        metadata = self._sm.get_metadata()
        if metadata:
            self.total = metadata.get('total', 0)

    def _dispatch(self, data):
        for service in data:
            # Update totals
            for entry in data[service]:
                self.total += entry['rating']['price']
        # Dispatch data to writing pipeline
        for backend in self._write_pipeline:
            backend.append(data, self.usage_start, self.usage_end)

    def get_timeframe(self, timeframe, timeframe_end=None):
        if not timeframe_end:
            timeframe_end = timeframe + self._period
        try:
            data = self._storage.get_time_frame(timeframe,
                                                timeframe_end,
                                                tenant_id=self._tenant_id)
        except storage.NoTimeFrame:
            return None
        return data

    def close(self):
        for writer in self._write_pipeline:
            writer.close()

    def _push_data(self):
        data = self.get_timeframe(self.usage_start, self.usage_end)
        if data:
            for timeframe in data:
                self._dispatch(timeframe['usage'])
            return True
        else:
            return False

    def _commit_data(self):
        for backend in self._write_pipeline:
            backend.commit()

    def reset_state(self):
        self._load_state_manager_data()
        self.usage_end = self._storage.get_state()
        self._update_state_manager_data()

    def restart_month(self):
        self._load_state_manager_data()
        month_start = ck_utils.get_month_start()
        self.usage_end = ck_utils.dt2ts(month_start)
        self._update_state_manager_data()

    def process(self):
        self._load_state_manager_data()
        storage_state = self._storage.get_state()
        if not self.usage_start:
            self.usage_start = storage_state
            self.usage_end = self.usage_start + self._period
        while storage_state > self.usage_start:
            if self._push_data():
                self._commit_data()
            self._update_state_manager_data()
            self._load_state_manager_data()
            storage_state = self._storage.get_state()
        self.close()



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import abc

import pecan
from pecan import rest
import six

from cloudkitty.common import policy
from cloudkitty.db import api as db_api
from cloudkitty import messaging


@six.add_metaclass(abc.ABCMeta)
class RatingProcessorBase(object):
    """Provides the Cloudkitty integration code to the rating processors.

    Every rating processor should subclass this and override at least
    module_name, description.

    config_controller can be left at None to use the default one.
    """

    module_name = None
    description = None
    config_controller = None
    hot_config = False

    @property
    def module_info(self):
        return {
            'name': self.module_name,
            'description': self.description,
            'hot_config': self.hot_config,
            'enabled': self.enabled,
            'priority': self.priority}

    def __init__(self, tenant_id=None):
        self._tenant_id = tenant_id

    @property
    def enabled(self):
        """Check if the module is enabled

        :returns: bool if module is enabled
        """
        api = db_api.get_instance()
        module_db = api.get_module_info()
        return module_db.get_state(self.module_name) or False

    @property
    def priority(self):
        """Get the priority of the module.

        """
        api = db_api.get_instance()
        module_db = api.get_module_info()
        return module_db.get_priority(self.module_name)

    def set_priority(self, priority):
        """Set the priority of the module.

        :param priority: (int) The new priority, the higher the number, the
        higher the priority.
        """
        api = db_api.get_instance()
        module_db = api.get_module_info()
        self.notify_reload()
        return module_db.set_priority(self.module_name, priority)

    def set_state(self, enabled):
        """Enable or disable a module.

        :param enabled: (bool) The state to put the module in.
        :return:  bool
        """
        api = db_api.get_instance()
        module_db = api.get_module_info()
        client = messaging.get_client().prepare(namespace='rating',
                                                fanout=True)
        if enabled:
            operation = 'enable_module'
        else:
            operation = 'disable_module'
        client.cast({}, operation, name=self.module_name)
        return module_db.set_state(self.module_name, enabled)

    def quote(self, data):
        """Compute rating informations from data.

        :param data: An internal CloudKitty dictionary used to describe
                     resources.
        :type data: dict(str:?)
        """
        return self.process(data)

    def nodata(self, begin, end):
        """Handle billing processing when no data has been collected.

        :param begin: Begin of the period.
        :param end: End of the period.
        """
        pass

    @abc.abstractmethod
    def process(self, data):
        """Add rating informations to data

        :param data: An internal CloudKitty dictionary used to describe
                     resources.
        :type data: dict(str:?)
        """

    @abc.abstractmethod
    def reload_config(self):
        """Trigger configuration reload

        """

    def notify_reload(self):
        client = messaging.get_client().prepare(namespace='rating',
                                                fanout=True)
        client.cast({}, 'reload_module', name=self.module_name)


class RatingRestControllerBase(rest.RestController):
    @pecan.expose()
    def _route(self, args, request):
        try:
            policy.authorize(request.context, 'rating:module_config', {})
        except policy.PolicyNotAuthorized as e:
            pecan.abort(403, six.text_type(e))

        return super(RatingRestControllerBase, self)._route(args, request)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_db.sqlalchemy import models
import sqlalchemy
from sqlalchemy.ext import declarative


Base = declarative.declarative_base()


class StateInfo(Base, models.ModelBase):
    """State

    """

    __tablename__ = 'states'

    name = sqlalchemy.Column(sqlalchemy.String(255),
                             primary_key=True)
    state = sqlalchemy.Column(
        sqlalchemy.BigInteger(),
        nullable=False)
    s_metadata = sqlalchemy.Column(sqlalchemy.Text(),
                                   nullable=True,
                                   default='')

    def __repr__(self):
        return ('<StateInfo[{name}]: '
                'state={state} metadata={metadata}>').format(
                    name=self.name,
                    state=self.state,
                    metadata=self.s_metadata)


class ModuleStateInfo(Base, models.ModelBase):
    """Module state info.

    """

    __tablename__ = 'modules_state'

    name = sqlalchemy.Column(sqlalchemy.String(255),
                             primary_key=True)
    state = sqlalchemy.Column(
        sqlalchemy.Boolean(),
        nullable=False,
        default=False)
    priority = sqlalchemy.Column(
        sqlalchemy.Integer(),
        default=1)

    def __repr__(self):
        return ('<ModuleStateInfo[{name}]: '
                'enabled={state}>').format(
                    name=self.name,
                    state=self.state)

    def as_dict(self):
        d = {}
        for c in self.__table__.columns:
            d[c.name] = self[c.name]
        return d


class ServiceToCollectorMapping(Base, models.ModelBase):
    """Collector module state.

    """

    __tablename__ = 'service_to_collector_mappings'

    service = sqlalchemy.Column(sqlalchemy.String(255),
                                primary_key=True)
    collector = sqlalchemy.Column(sqlalchemy.String(255),
                                  nullable=False)

    def __repr__(self):
        return ('<ServiceToCollectorMapping[{service}]: '
                'collector={collector}>').format(
                    service=self.service,
                    collector=self.collector)

    def as_dict(self):
        d = {}
        for c in self.__table__.columns:
            d[c.name] = self[c.name]
        return d



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""added tenant informations

Revision ID: 792b438b663
Revises: 17fd1b237aa3
Create Date: 2014-12-02 13:12:11.328534

"""

# revision identifiers, used by Alembic.
revision = '792b438b663'
down_revision = '17fd1b237aa3'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('rated_data_frames',
                  sa.Column('tenant_id', sa.String(length=32), nullable=True))
